"""
This module implements a stripes plot function that can create annual and monthly
stripes plots from given climate and weather data. There is a matplotlib and a 
plotly version. The streamlit app uses the plotly version by default.
The plotly plot is interactive and available in a German and English version, the
matplotlib version is static and only available in English.


"""
from typing import Tuple
import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns
import plotly.graph_objects as go


def yrplot_data(sr, ref=[1980, 2000], ag='sum', qa=0.95, cmxeq=True) -> Tuple[pd.DataFrame, float, float]:
    """
    Prepare the given data for monthly and annual stripes to be plotted in a heatmap.
    The data is grouped and the limits for plotting are returned, based on the quantile given.

    """
    # plot of heatmap with monthyl and annual stripes
    yrs = sr.index.year.unique()
    dummy = np.zeros((len(yrs), 14)) * np.nan
    dummy = pd.DataFrame(dummy)
    dummy.index = yrs
    dummy.columns = ['J', 'F', 'M', 'A', 'M', 'J', 'J', 'A', 'S', 'O', 'N', 'D', '', 'Year']

    for i in sr.index:
        dummy.iloc[i.year - yrs[0], i.month - 1] = sr.loc[i]
    for i in yrs:
        if ag == 'sum':
            dummy.iloc[i - yrs[0], 13] = sr.loc[sr.index.year == i].sum()
            cmx = 'coolwarm_r'
        elif ag == 'min':
            dummy.iloc[i - yrs[0], 13] = sr.loc[sr.index.year == i].min()
        elif ag == 'max':
            dummy.iloc[i - yrs[0], 13] = sr.loc[sr.index.year == i].max()
        else:  # ag == 'mean'
            dummy.iloc[i - yrs[0], 13] = sr.loc[sr.index.year == i].mean()

    if ref == None:
        pass
    else:
        refx = dummy.loc[ref[0]:ref[1]].mean(axis=0)
        dummy = dummy - refx

    if cmxeq:
        vxU = dummy.abs().quantile(qa).quantile(qa)
        vxL = -1. * vxU
    else:
        vxU = dummy.quantile(qa).quantile(qa)
        vxL = dummy.quantile(1. - qa).quantile(1. - qa)

    if ag == 'sum':
        dummy.iloc[:, 13] = dummy.iloc[:, 13] / 12

    return dummy, vxU, vxL


def _yrplot_matplotlib(sr, ref=[1980, 2000], ag='sum', qa=0.95, cbar_title='Temperature anomaly (K)', cmx='coolwarm', cmxeq=True, li=False):
    # prepare the data
    dummy, vxU, vxL = yrplot_data(sr, ref=ref, ag=ag, qa=qa, cmxeq=cmxeq)
    yrs = dummy.index.values
    # build the figure
    fig = plt.figure(figsize=(8,len(dummy)/15.))
    ax = sns.heatmap(dummy, cmap=cmx, vmin=vxL, vmax=vxU, cbar_kws={'label': cbar_title})

    if ref == None:
        pass
    else:
        ax.add_patch(
            plt.Rectangle((0, ref[0] - yrs[0]), 12, ref[1] - ref[0], fill=True, edgecolor='red', facecolor='gray', lw=2,
                          alpha=0.3, clip_on=False))
        ax.add_patch(
            plt.Rectangle((13, ref[0] - yrs[0]), 1, ref[1] - ref[0], fill=True, edgecolor='red', facecolor='gray', lw=2,
                          alpha=0.3, clip_on=False))
        ax.annotate('Reference period', (0.5, ref[1] - yrs[0] - 2), color='white', weight='bold', ha='left',
                    va='bottom', alpha=0.6)

    if type(li) == int:
        ax.axhline(li - yrs[0], color='k', ls='--', lw=1, alpha=0.5)
        ax.annotate(' >> observed', (12.5, li - yrs[0] - 0.5), color='k', ha='center', va='bottom', alpha=0.6,
                    rotation=90.)
        ax.annotate('modelled << ', (12.5, li - yrs[0] + 0.5), color='k', ha='center', va='top', alpha=0.6,
                    rotation=90.)
        # ax.add_patch(plt.Rectangle((0, li-yrs[0]), 12, 0, fill=False, edgecolor='k', ls='--', lw=1, alpha=0.5, clip_on=False))

    ax.set_ylabel('Year')
    ax.set_xlabel('Month          ')
    
    return fig


def _yrplot_plotly(sr, ref=[1980, 2000], ag='sum', qa=0.95, cbar_title='Temperature anomaly (K)', cmx='coolwarm', cmxeq=True, li=False, lang='en'):
    # prepare the data
    dummy, vxU, vxL = yrplot_data(sr, ref=ref, ag=ag, qa=qa, cmxeq=cmxeq)
    
    # use full names
    if lang == 'de':
        dummy.columns = ['Janua', 'Februar', 'März', 'April', 'Mai', 'Juni', 'Juli', 'August', 'September', 'Oktober', 'November', 'Dezember', '', 'Jahr']
        hv_tmpl = """<b>Jahr:</b> %{y}<br><b>Monat:</b> %{x}<br><b>Abweichung:</b> %{z:.1f} K"""
        ref_txt = f'<b>Reference Period</b><br>{ref[0]} - {ref[1]}'
    else:
        dummy.columns = ['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December', '', 'Year']
        hv_tmpl = """<b>Year:</b> %{y}<br><b>Month:</b> %{x}<br><b>Deviation:</b> %{z:.1f} K"""
        ref_txt = f'<b>Reference Period</b><br>{ref[0]} - {ref[1]}'
    
    fig = go.Figure()
    
    if cmx == 'coolwarm':
        lo = min(0, abs(vxL / dummy.min().min()))
        up = min(1, vxU / dummy.max().max())
        cmx = [
            [lo, '#8FBDD3'],
            [(up + lo) / 2 , '#E4D1B9'],
            [up, '#A97155'],
        ]
    # add the trace
    fig.add_trace(go.Heatmap(
        z=dummy,
        x=dummy.columns,
        y=dummy.index,
        colorscale=cmx,
        hovertemplate=hv_tmpl,
        showlegend=False,
        name=cbar_title
    ))

    # add reference window
    fig.add_shape(type='rect', x0=-0.5, y0=ref[0], x1=11.5, y1=ref[1], xref='x', fillcolor='rgba(255, 0,0,0.3)', line=dict(width=3, color='red'))
    fig.add_shape(type='rect', x0=12.5, y0=ref[0], x1=13.5, y1=ref[1], xref='x', fillcolor='rgba(255, 0,0,0.3)', line=dict(width=3, color='red'))
    fig.add_annotation(x=12, y=np.mean(ref), showarrow=False, text=ref_txt, textangle=-90, font=dict(size=16, color='red'))

    # add colorbar title
    fig.add_annotation(x=1.1, align='right', valign='top', xref='paper', yref='paper', xanchor='right', yanchor='middle', textangle=-90, text=cbar_title, showarrow=False, font=dict(size=16, color='black'))
    
    # update the layout
    fig.update_layout(
        template='plotly_white',
        height=min(len(dummy) * 15, 650),
        margin=dict(t=25, b=100, l=80, r=160, autoexpand=False),
        xaxis=dict(tickangle=45, title='Monat' if lang == 'de' else 'Month'),
        yaxis=dict(title='Jahr' if lang == 'de' else 'Year'),
    )

    # return
    return fig


def yrplot_hm(sr, ref=[1980, 2000], ag='sum', qa=0.95, cbar_title='Temperature anomaly (K)', cmx='coolwarm', cmxeq=True, li=False, lang='en', backend='plotly'):
    """
    @TODO: documentation
    """
    if backend.lower() == 'matplotlib':
        return _yrplot_matplotlib(sr, ref=ref, ag=ag, qa=qa, cbar_title=cbar_title, cmx=cmx, cmxeq=cmxeq, li=li)
    else:
        return _yrplot_plotly(sr, ref=ref, ag=ag, qa=qa, cbar_title=cbar_title, cmx=cmx, cmxeq=cmxeq, li=li, lang=lang)



import os
from ruins.core import Config

def get_test_config() -> Config:
    """Get a test config"""
    # overwrite some settings for unit tests
    args = dict(
        datapath=os.path.abspath(os.path.join(os.path.dirname(__file__), 'testdata')),
        debug=True
    )
    return Config(**args)



# -*- coding: utf-8 -*-

import streamlit as st
import pandas as pd
import numpy as np
import xarray as xr

import seaborn as sns
sns.set_style('whitegrid', {'grid.linestyle': u'--'})

from sdm import SDM

#import statsmodels.api as sm
#import altair as alt
import plotly.graph_objects as go
import plotly.express as px
from plotly.subplots import make_subplots
import matplotlib.pyplot as plt
from PIL import Image
from pathlib import Path

st.title('RUINS example for AGU')

def water_proj():
    #hsim = pd.read_csv('data/hsim.csv',index_col=0)
    #hsim.index = pd.to_datetime(hsim.index)
    hsim_collect = pd.read_csv('data/hsim_collect.csv', index_col=0)
    hsim_collect.index = pd.to_datetime(hsim_collect.index)
    all_vals = np.ravel(hsim_collect.values)[~np.isnan(np.ravel(hsim_collect.values))]

    perci = st.sidebar.slider('Adjust percentile of extreme events:', min_value=90., max_value=99.9999, value=99., key='perci')

    firstitem = True
    for i in hsim_collect.columns:
        hc = hsim_collect.loc[hsim_collect[i] > np.percentile(all_vals, perci), i].resample('1Y').count()
        hc.name = i
        if firstitem:
            h_count = hc
            firstitem = False
        else:
            h_count = pd.concat([h_count, hc], axis=1)

    fig = plt.figure(constrained_layout=True,figsize=(10,2.5))
    gs = fig.add_gridspec(1, 5)
    ax = fig.add_subplot(gs[0, :-1])
    ax1 = ax.twinx()
    ax2 = fig.add_subplot(gs[0, -1])

    all_vals = np.ravel(hsim_collect.values)[~np.isnan(np.ravel(hsim_collect.values))]
    sns.distplot(all_vals, ax=ax2)
    ax2.vlines(np.percentile(all_vals, perci), 0., 0.06, colors='red')
    ax2.text(np.percentile(all_vals, perci), 0.065, str(perci) + '%', c='red', ha='center')
    ax2.set_xlim(-3, np.percentile(all_vals, 99.98))

    for i in hsim_collect.columns:
        hsim_collect[i].plot(style='.', label='_nolegend_', c='gray', alpha=0.2, ax=ax)
        try:
            hsim_collect.loc[hsim_collect[i] > np.percentile(all_vals, perci), i].plot(style='.', label='_nolegend_',
                                                                                   c='red', alpha=0.2, ax=ax)
        except:
            pass

    (h_count.sum(axis=1) / len(h_count.columns)).rolling(7).mean().plot(ax=ax1, grid=False, ls='--', label='all')
    cc = [x for x in h_count.columns if x.split('.')[-1] == 'rcp45']
    (h_count[cc].sum(axis=1) / len(cc)).rolling(7).mean().plot(ax=ax1, grid=False, label='rcp45')

    cc = [x for x in h_count.columns if x.split('.')[-1] == 'rcp85']
    (h_count[cc].sum(axis=1) / len(cc)).rolling(7).mean().plot(ax=ax1, grid=False, label='rcp85')
    ax1.legend()
    ax.set_ylabel('Q (mm/day)')
    ax1.set_ylabel('days w/ Q>' + str(np.round(perci / 100., 3)) + ' percentile\n7 year rolling mean')
    st.pyplot()


ruinslogo = Image.open('RUINS_logo_small.png')

## sidebar header
st.sidebar.image(ruinslogo, caption='', use_column_width=True)
st.sidebar.header('Control Panel')
water_proj()



from ruins.apps import weather
from ruins.tests.util import get_test_config

from ruins.core import DataManager, Config

# create the test config
config = get_test_config()


def test_climate_indices():
    """Test only climate indices """
    # add the include_climate config
    config['include_climate'] = True
    dm = DataManager(**config)

    # run
    weather.climate_indices(dataManager=dm, config=config)


def test_climate_indi():
    """Test climate indi function"""
    conf = get_test_config()
    dm = DataManager(**conf)

    weath = dm['weather'].read()

    # make an arbitrary selection
    w = weath['coast'].sel(vars='Tmax').to_dataframe().dropna()
    w.columns = ['_', 'Tmax']

    # run
    weather.climate_indi(w, 'Ice days (Tmax < 0°C)')



import os
from ruins import __version__

def increment(which='patch'):
    """
    Increment the version number. 
    """
    parts = __version__.split('.')
    if which == 'patch':
        parts[2] = str(int(parts[2]) + 1)
    elif which == 'minor':
        parts[1] = str(int(parts[1]) + 1)
        parts[2] = '0'
    elif which == 'major':
        parts[0] = str(int(parts[0]) + 1)
        parts[1] = '0'
        parts[2] = '0'
    else:
        raise ValueError("Invalid version increment.")
    return '.'.join(parts)


def replace(which='patch'):
    """
    Increment the version number for RUINS.

    """
    # find the file
    path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'ruins', '__init__.py'))
    
    # read
    with open(path, 'r') as f:
        lines = f.readlines()
    
    # replace the version
    for i, line in enumerate(lines):
        if '__version__' in line:
            new_version = increment(which)
            lines[i] = f"__version__ = '{new_version}'\n"
            break
    
    # overwrite
    with open(path, 'w') as f:
        f.writelines(lines)
    
    print(new_version)  


if __name__ == '__main__':
    import fire
    fire.Fire(replace)


"""
Test the config object
"""
import os
import json
import pytest

from ruins import core


def test_default_config():
    """Check default paths"""
    # get the actual path
    basepath = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", '..'))
    datapath = os.path.join(basepath, 'data')

    # initialize config without arguments
    conf = core.Config()

    assert conf.datapath.lower() == datapath.lower()
    assert conf.basepath.lower() == basepath.lower()
    assert conf.debug is False


def test_config_overwrites():
    """Check that default config can be overwritten"""
    basepath = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", '..'))
    temppath = '/tmp/'

    conf = core.Config(datapath=temppath)

    assert conf.basepath.lower() == basepath.lower()
    assert conf.datapath.lower() == temppath.lower()


def test_from_json():
    """Build a config from file"""
    with open('test.json', 'w') as f:
        json.dump({'foo': 'bar'}, f)
    
    # build from config file
    conf = core.Config(path='test.json')

    assert conf['foo'] == 'bar'

    # remove test file
    os.remove('test.json')

def test_config_as_dict():
    """Test that the config behaves like a dict"""
    c = core.Config(foo='bar')

    # check custom configuration
    assert c['foo'] == 'bar'
    
    # check len and iter behavior
    i = 0
    for k in c:
        i += 1
    assert len(c) == i
    assert 'datapath' in c._keys

    # check default get behavior
    assert c.get('doesNotExist') is None
    assert c.get('doesNotExists', 'foobar') == 'foobar'


def test_config_key_error():
    """An unknown key should throw a key error"""
    c = core.Config()

    with pytest.raises(KeyError):
        c['doesNotExist']



import streamlit as st
import numpy as np

import xarray as xr # TODO  this should be covered by the DataManager
import pandas as pd # TODO this should be covered by the DataManager

from ruins.plotting import monthlyx
from ruins import components

# TODO: Build this into DataManager
def load_data(sel='Weather',regagg=None):
    #Read data from netcdf files and return xarray
    if sel == 'Weather':
        data = xr.load_dataset('data/weather.nc')
    elif sel == 'CORDEX':
        data = xr.load_dataset('data/cordex_coast.nc')
    else:
        if regagg == 'North Sea Coast':
            data = xr.load_dataset('data/cordex_coast.nc')
            data.filter_by_attrs(RCP=sel)

    return data


# TODO refactor the plots into the plotting module
def climate_explorer(w_topic: str):
    cliprojs = ["Global", "Regional"]
    cliproj = st.sidebar.radio("Climate Model Scaling:", options=cliprojs)

    # TODO: Re-implement this as a service
    # expl_md = read_markdown_file('explainer/climatescale.md')
    # st.sidebar.markdown(expl_md, unsafe_allow_html=True)

    if cliproj=='Regional':
        regaggs = ['North Sea Coast', 'Krummhörn',  'Niedersachsen', 'Inland']
        regagg = st.sidebar.selectbox('Spatial aggregation:', regaggs)

        if regagg=='North Sea Coast':
            climate = xr.load_dataset('data/cordex_coast.nc')
            climate.filter_by_attrs(RCP='rcp45')
        elif regagg=='Krummhörn':
            climate = xr.load_dataset('data/cordex_krummh.nc')

    st.subheader('Climate Model Comparison')
    crefs = ['Weather Data', 'all RCPs', 'RCP2.6', 'RCP4.5','RCP8.5']
    cref1 = st.sidebar.selectbox('Select first reference:', options=crefs)

    if cref1=='Weather Data':
        data1 = load_data('Weather')
        data1 = data1.coast
        drngx1 = (1980,2000)
    elif cref1 == 'all RCPs':
        data1 = load_data('CORDEX')
        drngx1 = (2050, 2070)
    elif cref1 == 'RCP2.6':
        data1 = load_data('rcp26','North Sea Coast')
        drngx1 = (2050, 2070)
    elif cref1 == 'RCP4.5':
        data1 = load_data('rcp45', 'North Sea Coast')
        drngx1 = (2050, 2070)
    elif cref1 == 'RCP8.5':
        data1 = load_data('rcp85','North Sea Coast')
        drngx1 = (2050, 2070)

    drng1 = [pd.to_datetime(data1.isel(time=0, vars=1).time.values).year, pd.to_datetime(data1.isel(time=-1, vars=1).time.values).year]
    datarng1 = st.sidebar.slider('Data range', drng1[0], drng1[1], drngx1, key='dr1')

    cref2 = st.sidebar.selectbox('Select second reference:', options=crefs)
    if cref2 == 'Weather Data':
        data2 = load_data('Weather')
        drngx2 = (1980, 2000)
    elif cref2 == 'all RCPs':
        data2 = load_data('CORDEX')
        drngx2 = (2050, 2070)
    elif cref2 == 'RCP2.6':
        data2 = load_data('rcp26','North Sea Coast')
        drngx2 = (2050, 2070)
    elif cref2 == 'RCP4.5':
        data2 = load_data('rcp45', 'North Sea Coast')
        drngx2 = (2050, 2070)
    elif cref2 == 'RCP8.5':
        data2 = load_data('rcp85','North Sea Coast')
        drngx2 = (2050, 2070)

    drng2 = [pd.to_datetime(data2.isel(time=0, vars=1).time.values).year, pd.to_datetime(data2.isel(time=-1, vars=1).time.values).year]
    datarng2 = st.sidebar.slider('Data range', drng2[0], drng2[1], drngx2, key='dr2')

    if w_topic == 'Warming':
        navi_vars = ['Maximum Air Temperature', 'Mean Air Temperature', 'Minimum Air Temperature']
        navi_var = st.sidebar.radio("Select variable:", options=navi_vars)
        if navi_var[:4] == 'Mini':
            vari = 'Tmin'
            afu = np.min
            ag = 'min'
        elif navi_var[:4] == 'Maxi':
            vari = 'Tmax'
            afu = np.max
            ag = 'max'
        else:
            vari = 'T'
            afu = np.mean
            ag = 'mean'

        dyp = data1.sel(vars=vari).to_dataframe().resample('1M').apply(afu)
        dyp = dyp.loc[(dyp.index.year>=datarng1[0]) & (dyp.index.year<datarng1[1]),dyp.columns[dyp.columns!='vars']]
        dyp2 = data2.sel(vars=vari).to_dataframe().resample('1M').apply(afu)
        dyp2 = dyp2.loc[(dyp2.index.year >= datarng2[0]) & (dyp2.index.year < datarng2[1]),dyp2.columns[dyp2.columns!='vars']]

        monthlyx(dyp, dyp2, 'Temperature (°C)', 'Monthly '+ag+' in Year ('+cref1+')', 'Monthly '+ag+' in Year ('+cref2+')')
        st.pyplot()


def main_app():
    st.header('Climate Projections Explorer')
    st.markdown('''In this section we add climate model projections to the table. The same variables and climate indices are used to explore the projections of different climate models and downscaling models. It is also possible to compare projections under different scenarios about the CO<sub>2</sub>-concentration pathways to observed weather and between different model projections.''',unsafe_allow_html=True)
    
    
    # TODO: refactor this
    topics = ['Warming', 'Weather Indices', 'Drought/Flood', 'Agriculture', 'Extreme Events', 'Wind Energy']
    
    # topic selector
    topic = components.topic_selector(topic_list=topics, container=st.sidebar)
    
    # TODO refactor this
    climate_explorer(topic)


if __name__ == '__main__':
    main_app()



# -*- coding: utf-8 -*-

import streamlit as st
import pandas as pd
import numpy as np
import xarray as xr

import seaborn as sns
sns.set_style('whitegrid', {'grid.linestyle': u'--'})

from sdm import SDM

#import statsmodels.api as sm
#import altair as alt
import plotly.graph_objects as go
import plotly.express as px
from plotly.subplots import make_subplots
import matplotlib.pyplot as plt
from PIL import Image
from pathlib import Path

st.title('RUINS climate data app')

# helper functions
#@st.cache(persist=True)
def load_alldata():
    weather = xr.load_dataset('data/weather.nc')
    climate = xr.load_dataset('data/cordex_coast.nc')

    # WARNING - bug fix for now:
    # 'HadGEM2-ES' model runs are problematic and will be removed for now
    # The issue is with the timestamp and requires revision of the ESGF reading routines
    kys = [s for s in list(climate.keys()) if 'HadGEM2-ES' not in s] #remove all entries of HadGEM2-ES (6 entries)
    climate = climate[kys]

    return weather, climate

weather, climate = load_alldata()


def load_data(sel='Weather',regagg=None):
    #Read data from netcdf files and return xarray
    if sel == 'Weather':
        data = xr.load_dataset('data/weather.nc')
    elif sel == 'CORDEX':
        data = xr.load_dataset('data/cordex_coast.nc')
    else:
        if regagg == 'North Sea Coast':
            data = xr.load_dataset('data/cordex_coast.nc')
            data.filter_by_attrs(RCP=sel)

    return data


def read_markdown_file(markdown_file):
    return Path(markdown_file).read_text()


def kde(data, cmdata='none', split_ts=1, cplot=True, eq_period=True):
    # plot of kde with stripes
    from sklearn.neighbors import KernelDensity
    cxx = ['#E69F00', '#009E73', '#0072B2', '#D55E00', '#CC79A7']
    cxx2 = ['#8c6bb1', '#810f7c']

    data = data[~np.isnan(data)]
    x_d = np.linspace(np.min(data) * 0.9, np.max(data) * 1.1, len(data))

    # instantiate and fit the KDE model
    kde = KernelDensity(bandwidth=1.0, kernel='gaussian')
    kde.fit(data[:, None])

    # score_samples returns the log of the probability density
    logprob = kde.score_samples(x_d[:, None])

    if cplot & (split_ts == 1):
        plt.fill_between(x_d, np.exp(logprob), alpha=0.4, facecolor='grey')

    lp = np.exp(logprob)
    xd = x_d

    if type(cmdata) != str:
        fig, (ax, cax, cax2) = plt.subplots(ncols=3, figsize=(10.5, 2.3), gridspec_kw={"width_ratios": [1, 0.02, 0.02]})

        x_d2 = np.linspace(np.min(cmdata) * 0.9, np.max(cmdata) * 1.1, len(cmdata))
        # instantiate and fit second KDE model
        kde2 = KernelDensity(bandwidth=1.0, kernel='gaussian')
        kde2.fit(cmdata[:, None])

        # score_samples returns the log of the probability density
        logprob2 = kde2.score_samples(x_d2[:, None])

        if cplot & (split_ts == 1):
            ax.fill_between(x_d2, np.exp(logprob2), alpha=0.4, facecolor='grey')

        lp2 = np.exp(logprob2)
        xd2 = x_d2
    else:
        fig, (ax, cax) = plt.subplots(ncols=2, figsize=(10.5, 2.3), gridspec_kw={"width_ratios": [1, 0.06]})

    if split_ts > 1:
        if eq_period:
            spliti = [0, len(data) - 40, len(data) - 20, len(data)]
        else:
            spliti = np.linspace(0, len(data), split_ts + 1).astype(int)

        for i in np.arange(split_ts):
            datax = data.iloc[spliti[i]:spliti[i + 1]]
            x_d = np.linspace(np.min(datax) * 0.9, np.max(datax) * 1.1, len(datax))

            try:
                # instantiate and fit the KDE model
                kde = KernelDensity(bandwidth=1.0, kernel='gaussian')
                kde.fit(datax[:, None])

                # score_samples returns the log of the probability density
                logprob = kde.score_samples(x_d[:, None])

                if cplot:
                    ax.fill_between(x_d, np.exp(logprob), alpha=0.4, facecolor=cxx[i],
                                    label='-'.join([str(datax.index.year.min()), str(datax.index.year.max())]))
            except:
                pass

    if type(cmdata) != str:
        # add climate model data
        csplit_ts = [2040, 2080]
        for i in np.arange(2):
            datax = cmdata.loc[(cmdata.index.year >= csplit_ts[i]) & (cmdata.index.year < csplit_ts[i] + 20)]
            x_d = np.linspace(np.min(datax) * 0.9, np.max(datax) * 1.1, len(datax))

            try:
                # instantiate and fit the KDE model
                kde = KernelDensity(bandwidth=1.0, kernel='gaussian')
                kde.fit(datax[:, None])

                # score_samples returns the log of the probability density
                logprob = kde.score_samples(x_d[:, None])

                if cplot:
                    ax.fill_between(x_d, np.exp(logprob), alpha=0.4, facecolor=cxx2[i],
                                    label='-'.join([str(datax.index.year.min()), str(datax.index.year.max())]))

            except:
                pass

    ax.legend(loc=1, ncol=5)

    if cplot:
        # cmap = plt.cm.get_cmap('cividis_r')
        cmap = plt.cm.get_cmap('viridis_r')
        colors = plt.cm.cividis_r(np.linspace(0, 1, len(data)))
        colorsx = cmap(np.arange(cmap.N))

        for i in np.arange(len(data)):
            ax.plot([data.iloc[i], data.iloc[i]], [0, np.max(lp) * 0.9], c=colors[i])

        labcb = 'Year'
        if type(cmdata) != str:
            cmap2 = plt.cm.get_cmap('plasma')
            colors2 = plt.cm.plasma(np.linspace(0, 1, len(cmdata)))
            colorsx2 = cmap2(np.arange(cmap2.N))

            for i in np.arange(len(cmdata)):
                ax.plot([cmdata.iloc[i], cmdata.iloc[i]], [0, np.max(lp) * 0.9], c=colors2[i])

            cbar2 = plt.colorbar(plt.cm.ScalarMappable(cmap=cmap2), cax=cax2, label=labcb,
                                 ticks=[0, 0.2, 0.4, 0.6, 0.8, 1], fraction=0.0027, anchor=(1.0, 0.1))
            cbar2.ax.set_yticklabels(
                np.round(np.linspace(cmdata.index.year.min(), cmdata.index.year.max(), 6)).astype(int).astype(str))
            labcb = ''

        cbar = plt.colorbar(plt.cm.ScalarMappable(cmap=cmap), cax=cax, label=labcb, ticks=[0, 0.2, 0.4, 0.6, 0.8, 1])
        cbar.ax.set_yticklabels(
            np.round(np.linspace(data.index.year.min(), data.index.year.max(), 6)).astype(int).astype(
                str))  # vertically oriented colorbar

    ax.set_ylabel('Occurrence (KDE)')

    return ax


def yrplot_hm(sr, ref=[1980, 2000], ag='sum', qa=0.95, cbar_title='Temperature anomaly (K)', cmx='coolwarm', cmxeq=True, li=False):
    # plot of heatmap with monthyl and annual stripes
    yrs = sr.index.year.unique()
    dummy = np.zeros((len(yrs), 14)) * np.nan
    dummy = pd.DataFrame(dummy)
    dummy.index = yrs
    dummy.columns = ['J', 'F', 'M', 'A', 'M', 'J', 'J', 'A', 'S', 'O', 'N', 'D', '', 'Year']

    for i in sr.index:
        dummy.iloc[i.year - yrs[0], i.month - 1] = sr.loc[i]
    for i in yrs:
        if ag == 'sum':
            dummy.iloc[i - yrs[0], 13] = sr.loc[sr.index.year == i].sum()
            cmx = 'coolwarm_r'
        elif ag == 'min':
            dummy.iloc[i - yrs[0], 13] = sr.loc[sr.index.year == i].min()
        elif ag == 'max':
            dummy.iloc[i - yrs[0], 13] = sr.loc[sr.index.year == i].max()
        else:  # ag == 'mean'
            dummy.iloc[i - yrs[0], 13] = sr.loc[sr.index.year == i].mean()

    if ref == None:
        pass
    else:
        refx = dummy.loc[ref[0]:ref[1]].mean(axis=0)
        dummy = dummy - refx

    if cmxeq:
        vxU = dummy.abs().quantile(qa).quantile(qa)
        vxL = -1. * vxU
    else:
        vxU = dummy.quantile(qa).quantile(qa)
        vxL = dummy.quantile(1. - qa).quantile(1. - qa)

    if ag == 'sum':
        dummy.iloc[:, 13] = dummy.iloc[:, 13] / 12

    plt.figure(figsize=(8,len(dummy)/15.))
    ax = sns.heatmap(dummy, cmap=cmx, vmin=vxL, vmax=vxU, cbar_kws={'label': cbar_title})

    if ref == None:
        pass
    else:
        ax.add_patch(
            plt.Rectangle((0, ref[0] - yrs[0]), 12, ref[1] - ref[0], fill=True, edgecolor='red', facecolor='gray', lw=2,
                          alpha=0.3, clip_on=False))
        ax.add_patch(
            plt.Rectangle((13, ref[0] - yrs[0]), 1, ref[1] - ref[0], fill=True, edgecolor='red', facecolor='gray', lw=2,
                          alpha=0.3, clip_on=False))
        ax.annotate('Reference period', (0.5, ref[1] - yrs[0] - 2), color='white', weight='bold', ha='left',
                    va='bottom', alpha=0.6)

    if type(li) == int:
        ax.axhline(li - yrs[0], color='k', ls='--', lw=1, alpha=0.5)
        ax.annotate(' >> observed', (12.5, li - yrs[0] - 0.5), color='k', ha='center', va='bottom', alpha=0.6,
                    rotation=90.)
        ax.annotate('modelled << ', (12.5, li - yrs[0] + 0.5), color='k', ha='center', va='top', alpha=0.6,
                    rotation=90.)
        # ax.add_patch(plt.Rectangle((0, li-yrs[0]), 12, 0, fill=False, edgecolor='k', ls='--', lw=1, alpha=0.5, clip_on=False))

    ax.set_ylabel('Year')
    ax.set_xlabel('Month          ')
    return


def monthlyx(dy, dyx=1, ylab='T (°C)', clab1='Monthly Mean in Year', clab2='Monthly Max in Year', pls='cividis_r'):
    cmap = plt.cm.get_cmap(pls)
    colors = cmap(np.linspace(0, 1, len(dy.index.year.unique())+1))
    colorsx = cmap(np.arange(cmap.N))

    idx1 = dy.index.year - dy.index.year.min()
    idx1m = dy.index.month

    if type(dyx) == int:
        pass
    else:
        idx2 = (dyx.index.year - dyx.index.year.min()).astype(int)
        idx2m = dyx.index.month

        cmap1 = plt.cm.get_cmap('gist_heat_r')
        colors1 = plt.cm.gist_heat_r(np.linspace(0, 1, len(dyx.index.year.unique()) + 1))
        colorsx1 = cmap1(np.arange(cmap1.N))

    for i in dy.columns:
        plt.scatter(idx1m + (np.random.rand(len(dy)) - 1.5), dy[i].values.astype(np.float), c=colors[idx1], alpha=0.6, s=2)

    if type(dyx) == int:
        pass
    else:
        for i in dyx.columns:
            plt.scatter(idx2m + (np.random.rand(len(dyx)) - 1.5), dyx[i].values.astype(np.float), c=colors1[idx2], alpha=0.6, s=2)

    cbar = plt.colorbar(plt.cm.ScalarMappable(cmap=cmap), label=clab1, ticks=[0, 0.2, 0.4, 0.6, 0.8, 1])
    cbar.ax.set_yticklabels(np.round(np.linspace(dy.index.year.min(), dy.index.year.max(), 6)).astype(int).astype(str))

    if type(dyx) == int:
        pass
    else:
        cbar1 = plt.colorbar(plt.cm.ScalarMappable(cmap=cmap1), label=clab2, ticks=[0, 0.2, 0.4, 0.6, 0.8, 1])
        cbar1.ax.set_yticklabels(
            np.round(np.linspace(dyx.index.year.min(), dyx.index.year.max(), 6)).astype(int).astype(str))

    plt.xticks([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
           ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'])
    plt.ylabel(ylab)
    return


def plt_map(sel='all',cm='none'):
    import plotly.graph_objs as go
    import plotly.express as px

    dummy = xr.open_dataset('data/CORDEXgrid.nc')
    dummy5 = xr.open_dataset('data/CMIP5grid.nc')
    stats = pd.read_csv('data/stats.csv', index_col=0)
    stats['ms'] = 15.
    stats['color'] = 'gray'

    mapbox_access_token = 'pk.eyJ1IjoiY29qYWNrIiwiYSI6IkRTNjV1T2MifQ.EWzL4Qk-VvQoaeJBfE6VSA'
    px.set_mapbox_access_token(mapbox_access_token)

    nodexy = pd.DataFrame([dummy.lon.values.ravel(), dummy.lat.values.ravel()]).T
    nodexy.columns = ['lon', 'lat']
    nodexy['hov'] = 'CORDEX grid'

    def lin_grid(fig, dummy, chex='#d95f0e', opc=0.7):
        x1, x2 = np.shape(dummy.lon)
        lond = np.diff(dummy.lon.values)
        latd = np.diff(dummy.lat.values)
        lond1 = np.diff(dummy.lon.values, axis=0)
        latd1 = np.diff(dummy.lat.values, axis=0)

        for i in np.arange(x2)[1:-1]:
            fig.add_trace(go.Scattermapbox(
                mode='lines',
                lon=dummy.lon.values[:, i] - 0.5 * lond[:, i - 1],
                lat=dummy.lat.values[:, i] - 0.5 * latd[:, i - 1],
                line={'color': chex, 'width': 1},
                hoverinfo='skip',
                opacity=opc))

        fig.add_trace(go.Scattermapbox(
            mode='lines',
            lon=dummy.lon.values[:, i] + 0.5 * lond[:, i - 1],
            lat=dummy.lat.values[:, i] + 0.5 * latd[:, i - 1],
            line={'color': chex, 'width': 1},
            hoverinfo='skip',
            opacity=opc))

        for i in np.arange(x1)[1:-1]:
            fig.add_trace(go.Scattermapbox(
                mode='lines',
                lon=dummy.lon.values[i, :] - 0.5 * lond1[i - 1, :],
                lat=dummy.lat.values[i, :] - 0.5 * latd1[i - 1, :],
                line={'color': chex, 'width': 1},
                hoverinfo='skip',
                opacity=opc))

        fig.add_trace(go.Scattermapbox(
            mode='lines',
            lon=dummy.lon.values[i, :] + 0.5 * lond1[i - 1, :],
            lat=dummy.lat.values[i, :] + 0.5 * latd1[i - 1, :],
            line={'color': chex, 'width': 1},
            hoverinfo='skip',
            opacity=opc))

        return fig

    def add_stats(sel='all'):


        if sel == 'all':
            ac = '#2c7fb8'
            stats1 = stats
        else:
            ac = 'gray'

            fig.add_trace(go.Scattermapbox(
                lat=stats.lat,
                lon=stats.lon,
                mode='markers',
                marker=go.scattermapbox.Marker(
                    size=10,
                    color=ac,
                    opacity=0.8
                ),
                text=stats['Station name']
            ))

            ac = '#df65b0'
            if sel == 'krummhoern':
                stats1 = stats.loc[stats.krummhoern == True]
            elif sel == 'coast':
                stats1 = stats.loc[stats.coast == True]
            elif sel == 'niedersachsen':
                stats1 = stats.loc[stats.niedersachsen == True]
            elif sel == 'inland':
                stats1 = stats.loc[stats.inland == True]
            else:
                try:
                    stats1 = pd.DataFrame(stats.loc[sel]).T
                except:
                    ac = 'gray'
                    stats1 = pd.DataFrame(stats.loc['Norderney']).T

        fig.add_trace(go.Scattermapbox(
            lat=stats1.lat,
            lon=stats1.lon,
            mode='markers',
            marker=go.scattermapbox.Marker(
                size=10,
                color=ac,
                opacity=0.8
            ),
            text=stats1['Station name']
        ))

        return fig

    def add_cmpx(cm):
        maskcordex_coast = [[10, 4], [11, 4], [11, 5], [11, 6], [11, 7], [11, 8], [10, 8], [9, 8], [8, 8], [8, 9],
                            [8, 10], [9, 10], [10, 10], [10, 11], [11, 12], [12, 12], [12, 13], [12, 14], [13, 14],
                            [14, 14], [15, 14], [16, 14], [17, 14], [18, 14], [19, 15], [20, 15], [21, 14], [22, 14],
                            [22, 13],
                            [9, 4], [10, 5], [10, 6], [10, 7], [10, 8], [9, 7], [8, 7], [7, 7], [7, 8], [7, 9], [9, 9],
                            [10, 12], [11, 13], [11, 14], [14, 15], [15, 15], [16, 15], [17, 15], [18, 15], [19, 16],
                            [20, 16], [21, 15], [22, 15],
                            [11, 4], [12, 4], [12, 5], [12, 6], [12, 7], [12, 8], [11, 10], [11, 11], [13, 12],
                            [13, 13], [14, 13], [15, 13], [16, 13], [17, 13], [18, 13], [19, 14], [20, 14], [21, 13]]
        for cc in maskcordex_coast:
            fig.add_trace(go.Scattermapbox(
                lat=[dummy.lat.values[tuple(cc)]],
                lon=[dummy.lon.values[tuple(cc)]],
                mode='markers',
                marker=go.scattermapbox.Marker(
                    size=8,
                    color='#fec44f',
                    opacity=0.8
                ),
                hoverinfo='skip'
            ))
        return fig

    # fig = go.Scattermapbox(lon = nodexy.lon, lat= nodexy.lat, mode='markers', marker_symbol='square', marker_size=15)
    #fig = px.scatter_mapbox(nodexy, lat='lat', lon='lon', center={'lat': 53.0, 'lon': 8.3}, zoom=5, opacity=0.1, hover_data=['hov'])
    fig = px.scatter_mapbox(stats, lat='lat', lon='lon', center={'lat': 53.0, 'lon': 8.6}, zoom=5, size='ms', opacity=0.8, color='color', hover_data=['Station name', 'lat', 'lon'], size_max=10)
    if cm != 'none':
        fig = lin_grid(fig, dummy)
        fig = lin_grid(fig, dummy5, '#2c7fb8')
        fig = add_cmpx(cm)
    fig = add_stats(sel)
    fig.update_layout(showlegend=False,width=300, height=350,margin=dict(l=10, r=10, b=10, t=10))  # ,center={'lat':54.0,'lon':8.3}, zoom=7)
    st.sidebar.plotly_chart(fig)
    return


def applySDM(wdata, data, meth='rel', cdf_threshold=0.9999999, lower_limit=0.1):
    '''apply structured distribution mapping to climate data and return unbiased version of dataset'''
    from sdm import SDM
    data_ub = data.copy()

    for k in data_ub.columns:
        data_col = data_ub[k].dropna()
        overlapx = pd.concat(
            [wdata.loc[data_col.index[0]:wdata.index[-1]], data_col.loc[data_col.index[0]:wdata.index[-1]]], axis=1)
        overlapx.columns = ['obs', 'cm']
        overlapx = overlapx.dropna()
        try:
            data_ub[k] = SDM(overlapx.obs, overlapx.cm, data_col, meth, cdf_threshold, lower_limit)
        except:
            data_ub[k] = data_ub[k] * np.nan

    data_ub[data_ub == 0.0000000] = np.nan
    data_ub = data_ub.loc[data_ub.index[0]:pd.to_datetime('2099-12-31 23:59:59')]

    return data_ub


def climate_indi(ts, indi='Summer days (Tmax ≥ 25°C)'):
    '''
    Calculate climate indicator days.
    Input time series of meteorological data
    '''
    if pd.infer_freq(ts.index) != 'D':
        print('Please provide daily data.')
        return

    if indi == 'Summer days (Tmax ≥ 25°C)':  # summer days
        return (ts.Tmax >= 25.).groupby(ts.index.year).sum()
    elif indi == 'Ice days (Tmax < 0°C)':  # ice days
        return (ts.Tmax < 0.).groupby(ts.index.year).sum()
    elif indi == 'Frost days (Tmin < 0°C)':  # frost days
        return (ts.Tmin < 0.).groupby(ts.index.year).sum()
    elif indi == 'Hot days (Tmax ≥ 30°C)':  # hot days
        return (ts.Tmax >= 30.).groupby(ts.index.year).sum()
    elif indi == 'Tropic nights (Tmin ≥ 20°C)':  # tropic night
        return (ts.Tmin >= 20.).groupby(ts.index.year).sum()
    elif indi == 'Rainy days (Precip ≥ 1mm)':  # rainy days
        return (ts.Prec >= 1.).groupby(ts.index.year).sum()
    else:
        print('Nothing calculated.')
        return


def climate_indices(stati='coast',cliproj=True):
    cindi = ['Ice days (Tmax < 0°C)', 'Frost days (Tmin < 0°C)', 'Summer days (Tmax ≥ 25°C)', 'Hot days (Tmax ≥ 30°C)','Tropic nights (Tmin ≥ 20°C)', 'Rainy days (Precip ≥ 1mm)']
    ci_topic = st.selectbox('Select Index:', cindi)

    if ci_topic == 'Rainy days (Precip ≥ 1mm)':
        vari = 'Prec'
        meth = 'rel'
    elif (ci_topic == 'Frost days (Tmin < 0°C)') | (ci_topic == 'Tropic nights (Tmin ≥ 20°C)'):
        vari = 'Tmin'
        meth = 'abs'
    else:
        vari = 'Tmax'
        meth = 'abs'

    w1 = weather[stati].sel(vars=vari).to_dataframe().dropna()
    w1.columns = ['bla', vari]

    plt.figure(figsize=(10,2.5))
    wi = climate_indi(w1, ci_topic).astype(int)
    wi.plot(style='.', color='steelblue', label='Coast weather')
    wi.rolling(10, center=True).mean().plot(color='steelblue', label='Rolling mean\n(10 years)')

    if cliproj:
        c1 = climate.sel(vars=vari).to_dataframe()
        c1 = c1[c1.columns[c1.columns != 'vars']]
        c2 = applySDM(w1[vari], c1, meth=meth)

        firstitem = True
        for j in c2.columns:
            dummyt = pd.DataFrame(c2[j])
            dummyt.columns = [vari]
            ci = pd.DataFrame(climate_indi(dummyt, ci_topic).astype(int))
            ci.columns = [j]

            if firstitem:
                ci[j].plot(style='.', color='gray', alpha=0.1, label='all climate projections')
                cid = ci
                firstitem = False
            else:
                ci[j].plot(style='.', color='gray', alpha=0.1, label='_')
                cid = pd.concat([cid, ci], axis=1)

        rcpx = np.array([x[-5:] for x in cid.columns.values])
        for n in np.unique(rcpx):
            nx = n
            if n == np.unique(rcpx)[-1]:
                nx = n + '\n(Rolling of 5 years)'
            cid[cid.columns[rcpx == n]].mean(axis=1).rolling(5, center=True).mean().plot(label='Mean of ' + nx)

    plt.legend(ncol=2)
    plt.ylabel('Number of days')
    plt.title(ci_topic)
    st.pyplot()

    if ci_topic == 'Ice days (Tmax < 0°C)':
        st.markdown('''Number of days in one year which persistently remain below 0°C air temperature.''')
    elif ci_topic == 'Frost days (Tmin < 0°C)':
        st.markdown('''Number of days in one year which reached below 0°C air temperature.''')
    elif ci_topic == 'Summer days (Tmax ≥ 25°C)':
        st.markdown('''Number of days in one year which reached or exceeded 25°C air temperature.''')
    elif ci_topic == 'Hot days (Tmax ≥ 30°C)':
        st.markdown('''Number of days in one year which reached or exceeded 30°C air temperature.''')
    elif ci_topic == 'Tropic nights (Tmin ≥ 20°C)':
        st.markdown('''Number of days in one year which persistently remained above 20°C air temperature.''')
    elif ci_topic == 'Rainy days (Precip ≥ 1mm)':
        st.markdown('''Number of days in one year which received at least 1 mm precipitation.''')
    return



def weather_explorer():
    #weather = load_data('Weather')

    #aspects = ['Annual', 'Monthly', 'Season']
    #w_aspect = st.sidebar.selectbox('Temporal aggegate:', aspects)

    #cliproj = st.sidebar.checkbox('add climate projections',False)

    statios = list(weather.keys())
    stat1 = st.selectbox('Select station/group (see map in sidebar for location):', statios)

    aspects = ['Annual', 'Monthly']  # , 'Season']
    w_aspect = st.selectbox('Select temporal aggegate:', aspects)

    cliproj = st.checkbox('add climate projections (for coastal region)',False)
    if cliproj:
        plt_map(stat1, 'CORDEX')
        st.sidebar.markdown(
            '''Map with available stations (<span style="color:blue">blue dots</span>) and selected reference station (<span style="color:magenta">magenta highlight</span>). The climate model grid is given in <span style="color:orange">orange</span> with the selected references as filled dots.''',
            unsafe_allow_html=True)
    else:
        plt_map(stat1)
        st.sidebar.markdown(
            '''Map with available stations (<span style="color:blue">blue dots</span>) and selected reference station (<span style="color:magenta">magenta highlight</span>).''',
            unsafe_allow_html=True)

    if w_topic == 'Warming':

        navi_vars = ['Maximum Air Temperature', 'Mean Air Temperature', 'Minimum Air Temperature']
        navi_var = st.sidebar.radio("Select variable:", options=navi_vars)
        if navi_var[:4] == 'Mini':
            vari = 'Tmin'
            afu = np.min
            ag = 'min'
        elif navi_var[:4] == 'Maxi':
            vari = 'Tmax'
            afu = np.max
            ag = 'max'
        else:
            vari = 'T'
            afu = np.mean
            ag = 'mean'

        if w_aspect == 'Annual':
            wdata = weather[stat1].sel(vars=vari).resample(time='1Y').apply(afu).to_dataframe()[stat1]
            wdata = wdata[~np.isnan(wdata)]
            allw = weather.sel(vars=vari).resample(time='1Y').apply(afu).to_dataframe().iloc[:, 1:]

            dataLq = float(np.floor(allw.min().quantile(0.22)))
            datamin = float(np.min([dataLq, np.round(allw.min().min(), 1)]))
            if cliproj:
                rcps = ['rcp26', 'rcp45', 'rcp85']
                rcp = st.selectbox(
                    'RCP (Mean over all projections will be shown. For more details go to section "Climate Projections"):',
                    rcps)

                data = climate.filter_by_attrs(RCP=rcp).sel(vars=vari).resample(time='1Y').apply(afu).to_dataframe()
                data = data[data.columns[data.columns != 'vars']]
                data_ub = applySDM(wdata, data, meth='abs')

                dataUq = float(np.ceil(data_ub.max().quantile(0.76)))
                datamax = float(np.max([dataUq, np.round(data_ub.max().max(), 1)]))
            else:
                dataUq = float(np.ceil(allw.max().quantile(0.76)))
                datamax = float(np.max([dataUq,np.round(allw.max().max(), 1)]))

            datarng = st.slider('Adjust data range on x-axis of plot:', min_value=datamin, max_value=datamax, value=(dataLq, dataUq), step=0.1, key='drangew')

            if cliproj:
                ax = kde(wdata, data_ub.mean(axis=1), split_ts=3)
            else:
                ax = kde(wdata, split_ts=3)

            ax.set_title(stat1 + ' Annual ' + navi_var)
            ax.set_xlabel('T (°C)')
            ax.set_xlim(datarng[0],datarng[1])
            st.pyplot()

            sndstat = st.checkbox('Show second station for comparison')

            if sndstat:
                stat2 = st.selectbox('Select second station:', [x for x in statios if x != stat1])
                wdata2 = weather[stat2].sel(vars=vari).resample(time='1Y').apply(afu).to_dataframe()[stat2]

                ax2 = kde(wdata2, split_ts=3)
                ax2.set_title(stat2 + ' Annual ' + navi_var)
                ax2.set_xlabel('T (°C)')
                ax2.set_xlim(datarng[0],datarng[1])
                st.pyplot()

            expl_md = read_markdown_file('explainer/stripes.md')
            st.markdown(expl_md, unsafe_allow_html=True)

        elif w_aspect == 'Monthly':
            wdata = weather[stat1].sel(vars=vari).resample(time='1M').apply(afu).to_dataframe()[stat1]
            wdata = wdata[~np.isnan(wdata)]
            ref_yr = st.slider('Reference period for anomaly calculation:', min_value=int(wdata.index.year.min()), max_value=2020,value=(max(1980, int(wdata.index.year.min())), 2000))

            if cliproj:
                rcps = ['rcp26', 'rcp45', 'rcp85']
                rcp = st.selectbox('RCP (Mean over all projections will be shown. For more details go to section "Climate Projections"):', rcps)

                data = climate.filter_by_attrs(RCP=rcp).sel(vars=vari).resample(time='1M').apply(afu).to_dataframe()
                data = data[data.columns[data.columns != 'vars']]

                #ub = st.sidebar.checkbox('Apply SDM bias correction',True)
                ub = True # simplify here and automatically apply bias correction

                if ub:
                    data_ub = applySDM(wdata, data, meth='abs')
                    yrplot_hm(pd.concat([wdata.loc[wdata.index[0]:data.index[0] - pd.Timedelta('1M')], data_ub.mean(axis=1)]),ref_yr, ag, li=2006)
                else:
                    yrplot_hm(pd.concat([wdata.loc[wdata.index[0]:data.index[0] - pd.Timedelta('1M')], data.mean(axis=1)]), ref_yr, ag, li=2006)

                plt.title(stat1 + ' ' + navi_var + ' anomaly to ' + str(ref_yr[0]) + '-' + str(ref_yr[1]))
                st.pyplot()



            else:
                yrplot_hm(wdata,ref_yr,ag)
                plt.title(stat1 + ' ' + navi_var + ' anomaly to ' + str(ref_yr[0]) + '-' + str(ref_yr[1]))
                st.pyplot()

                sndstat = st.checkbox('Compare to a second station?')

                if sndstat:
                    stat2 = st.selectbox('Select second station:', [x for x in statios if x != stat1])
                    data2 = weather[stat2].sel(vars=vari).resample(time='1M').apply(afu).to_dataframe()[stat2]
                    data2 = data2[~np.isnan(data2)]

                    ref_yr2 = list(ref_yr)
                    if ref_yr2[1]<data2.index.year.min():
                        ref_yr2[0] = data2.index.year.min()
                        ref_yr2[1] = ref_yr2[0]+10
                    if ref_yr2[0]<data2.index.year.min():
                        ref_yr2[0] = data2.index.year.min()
                        if ref_yr2[1] - ref_yr2[0] < 10:
                            ref_yr2[1] = ref_yr2[0] + 10

                    yrplot_hm(data2, ref_yr2, ag)
                    plt.title(stat2 + ' ' + navi_var + ' anomaly to ' + str(ref_yr2[0]) + '-' + str(ref_yr2[1]))
                    st.pyplot()

            expl_md = read_markdown_file('explainer/stripes_m.md')
            st.markdown(expl_md, unsafe_allow_html=True)

    elif w_topic == 'Weather Indices':
        climate_indices(stat1,cliproj)

    if cliproj:
        st.markdown(
            '''RCPs are scenarios about possible greenhouse gas concentrations by the year 2100. RCP2.6 is a world in which little further greenhouse gasses are emitted -- similar to the Paris climate agreement from 2015. RCP8.5 was intendent to explore a rather risky, more worst-case future with further increased emissions. RCP4.5 is one candidate of a more moderate greenhouse gas projection, which might be more likely to resemble a realistic situation. It is important to note that the very limited differentiation between RCP scenarios have been under debate for several years. One outcome is the definition of Shared Socioeconomic Pathways (SSPs) for which today, however, there are not very many model runs awailable. For more information, please check with the [Climatescenario Primer](https://climatescenarios.org/primer/), [CarbonBrief](https://www.carbonbrief.org/explainer-how-shared-socioeconomic-pathways-explore-future-climate-change) and this [NatureComment](https://www.nature.com/articles/d41586-020-00177-3)''',
            unsafe_allow_html=True)


def climate_explorer():
    cliprojs = ["Global", "Regional"]
    cliproj = st.sidebar.radio("Climate Model Scaling:", options=cliprojs)

    expl_md = read_markdown_file('explainer/climatescale.md')
    st.sidebar.markdown(expl_md, unsafe_allow_html=True)

    if cliproj=='Regional':
        regaggs = ['North Sea Coast', 'Krummhörn',  'Niedersachsen', 'Inland']
        regagg = st.sidebar.selectbox('Spatial aggregation:', regaggs)

        if regagg=='North Sea Coast':
            climate = xr.load_dataset('data/cordex_coast.nc')
            climate.filter_by_attrs(RCP='rcp45')
        elif regagg=='Krummhörn':
            climate = xr.load_dataset('data/cordex_krummh.nc')

    st.subheader('Climate Model Comparison')
    crefs = ['Weather Data', 'all RCPs', 'RCP2.6', 'RCP4.5','RCP8.5']
    cref1 = st.sidebar.selectbox('Select first reference:', options=crefs)

    if cref1=='Weather Data':
        data1 = load_data('Weather')
        data1 = data1.coast
        drngx1 = (1980,2000)
    elif cref1 == 'all RCPs':
        data1 = load_data('CORDEX')
        drngx1 = (2050, 2070)
    elif cref1 == 'RCP2.6':
        data1 = load_data('rcp26','North Sea Coast')
        drngx1 = (2050, 2070)
    elif cref1 == 'RCP4.5':
        data1 = load_data('rcp45', 'North Sea Coast')
        drngx1 = (2050, 2070)
    elif cref1 == 'RCP8.5':
        data1 = load_data('rcp85','North Sea Coast')
        drngx1 = (2050, 2070)

    drng1 = [pd.to_datetime(data1.isel(time=0, vars=1).time.values).year, pd.to_datetime(data1.isel(time=-1, vars=1).time.values).year]
    datarng1 = st.sidebar.slider('Data range', drng1[0], drng1[1], drngx1, key='dr1')

    cref2 = st.sidebar.selectbox('Select second reference:', options=crefs)
    if cref2 == 'Weather Data':
        data2 = load_data('Weather')
        drngx2 = (1980, 2000)
    elif cref2 == 'all RCPs':
        data2 = load_data('CORDEX')
        drngx2 = (2050, 2070)
    elif cref2 == 'RCP2.6':
        data2 = load_data('rcp26','North Sea Coast')
        drngx2 = (2050, 2070)
    elif cref2 == 'RCP4.5':
        data2 = load_data('rcp45', 'North Sea Coast')
        drngx2 = (2050, 2070)
    elif cref2 == 'RCP8.5':
        data2 = load_data('rcp85','North Sea Coast')
        drngx2 = (2050, 2070)

    drng2 = [pd.to_datetime(data2.isel(time=0, vars=1).time.values).year, pd.to_datetime(data2.isel(time=-1, vars=1).time.values).year]
    datarng2 = st.sidebar.slider('Data range', drng2[0], drng2[1], drngx2, key='dr2')

    if w_topic == 'Warming':
        navi_vars = ['Maximum Air Temperature', 'Mean Air Temperature', 'Minimum Air Temperature']
        navi_var = st.sidebar.radio("Select variable:", options=navi_vars)
        if navi_var[:4] == 'Mini':
            vari = 'Tmin'
            afu = np.min
            ag = 'min'
        elif navi_var[:4] == 'Maxi':
            vari = 'Tmax'
            afu = np.max
            ag = 'max'
        else:
            vari = 'T'
            afu = np.mean
            ag = 'mean'

        dyp = data1.sel(vars=vari).to_dataframe().resample('1M').apply(afu)
        dyp = dyp.loc[(dyp.index.year>=datarng1[0]) & (dyp.index.year<datarng1[1]),dyp.columns[dyp.columns!='vars']]
        dyp2 = data2.sel(vars=vari).to_dataframe().resample('1M').apply(afu)
        dyp2 = dyp2.loc[(dyp2.index.year >= datarng2[0]) & (dyp2.index.year < datarng2[1]),dyp2.columns[dyp2.columns!='vars']]

        monthlyx(dyp, dyp2, 'Temperature (°C)', 'Monthly '+ag+' in Year ('+cref1+')', 'Monthly '+ag+' in Year ('+cref2+')')
        st.pyplot()


def cropmodel(data,crop='wheat',rcp='rcp85',name='croprunname'):
    import get_climate_data.cli_crop as clc
    #read co2 concentrations
    if rcp == 'rcp85':
        CO2 = pd.read_csv('data/RCP85_MIDYEAR_CONCENTRATIONS.DAT', skiprows=38, delim_whitespace=True, index_col=0).CO2EQ
    elif rcp == 'rcp45':
        CO2 = pd.read_csv('data/RCP45_MIDYEAR_CONCENTRATIONS.DAT', skiprows=38, delim_whitespace=True, index_col=0).CO2EQ
    else:
        CO2 = pd.read_csv('data/RCP45_MIDYEAR_CONCENTRATIONS.DAT', skiprows=38, delim_whitespace=True, index_col=0).CO2EQ * 0. + 400.

    if crop == 'maize':
        x = np.array([9.375e+00, 3.198e+01, 1.973e+00, 8.700e+01, 1.144e+01, 3.630e+01, 7.260e-02, 1.237e+00, 2.180e+03, 1.501e-01, 5.230e-01, 5.678e+01, 6.970e+02])
    elif crop == 'meadow':
        x = np.array([6.543e+00, 1.238e+01, 1.029e+00, 8.706e+01, 1.510e+01, 3.253e+01, 1.199e+00, 1.535e+03, 7.784e+03, 6.530e+03, 8.030e+03, 8.092e+03, 5.884e+03])
    elif crop == 'wheat':
        x = np.array([1.257e+00, 1.276e+01, 1.101e+00, 1.010e+02, 2.578e+01, 2.769e+01, 3.416e-01, 4.940e-01, 1.906e+03, 1.921e-01, 4.595e-01, 6.066e+01, 5.360e+02])
    else:
        print('ERROR: Crop not specified with parameters.')

    yields = clc.cli_SC(data, x, CO2, nme=name)

    return yields


def ub_climate(cdata, wdata, ub=True):
    varis = ['T', 'Tmax', 'Tmin', 'aP', 'Prec', 'RH', 'Rs', 'u2', 'EToHG']
    firstitem = True
    for vari in varis:
        data = cdata.sel(vars=vari).to_dataframe()
        data = data[data.columns[data.columns != 'vars']]

        if (vari == 'T') | (vari == 'Tmax') | (vari == 'Tmin'):
            meth = 'abs'
        else:
            meth = 'rel'

        if ub:
            wdatax = wdata.sel(vars=vari).to_dataframe().iloc[:, -1].dropna()
            data_ub = applySDM(wdatax, data, meth=meth)

        else:
            data_ub = data

        data_ubx = data_ub.mean(axis=1)
        data_ubx.columns = [vari]

        if firstitem:
            data_ubc = data_ubx
            firstitem = False
        else:
            data_ubc = pd.concat([data_ubc, data_ubx], axis=1)

    data_ubc.columns = varis
    return data_ubc


def get_turbine(pfi, plotit=False):
    dummy = pd.read_csv(pfi)
    v_start = dummy.loc[3].astype(np.float).values[0]
    v_max = dummy.loc[2].astype(np.float).values[0]
    P_v = dummy.loc[4:28].astype(np.float)
    P_v.index = np.arange(int(np.ceil(v_max)) + 1)[1:]
    if plotit:
        P_v.plot()
    return [P_v, v_start, v_max]


def P_wind(wind, pfi):
    [P_v, v_start, v_max] = get_turbine(pfi, False)

    # stephours = (wind.index[2] - wind.index[1]).days * 24
    # stephours = 1.

    def interp(val):
        if (val >= v_max) | (val < v_start):
            return 0.
        elif ~np.isnan(val):
            if np.ceil(val) == np.floor(val):
                return P_v.loc[int(np.floor(val))].values[0]
            else:
                x1 = P_v.loc[int(np.floor(val))].values[0]
                x2 = P_v.loc[int(np.ceil(val))].values[0]
                return ((val - np.floor(val)) * x2 + (1. - (val - np.floor(val))) * x1)
        else:
            return np.nan

    ip_vec = np.vectorize(interp)

    def get_windpower(u2):
        val = np.fmax(-1. * np.cos(np.arange(0., 2 * np.pi, 0.27)) + u2, 0.)
        return np.sum(ip_vec(val))

    return wind.apply(get_windpower)

def water_proj():
    #hsim = pd.read_csv('data/hsim.csv',index_col=0)
    #hsim.index = pd.to_datetime(hsim.index)
    hsim_collect = pd.read_csv('data/hsim_collect.csv', index_col=0)
    hsim_collect.index = pd.to_datetime(hsim_collect.index)
    all_vals = np.ravel(hsim_collect.values)[~np.isnan(np.ravel(hsim_collect.values))]

    perci = st.slider('Adjust percentile of extreme events:', min_value=90., max_value=99.9999, value=99., key='perci')

    firstitem = True
    for i in hsim_collect.columns:
        hc = hsim_collect.loc[hsim_collect[i] > np.percentile(all_vals, perci), i].resample('1Y').count()
        hc.name = i
        if firstitem:
            h_count = hc
            firstitem = False
        else:
            h_count = pd.concat([h_count, hc], axis=1)

    fig = plt.figure(constrained_layout=True,figsize=(10,2.5))
    gs = fig.add_gridspec(1, 5)
    ax = fig.add_subplot(gs[0, :-1])
    ax1 = ax.twinx()
    ax2 = fig.add_subplot(gs[0, -1])

    all_vals = np.ravel(hsim_collect.values)[~np.isnan(np.ravel(hsim_collect.values))]
    sns.distplot(all_vals, ax=ax2)
    ax2.vlines(np.percentile(all_vals, perci), 0., 0.06, colors='red')
    ax2.text(np.percentile(all_vals, perci), 0.065, str(perci) + '%', c='red', ha='center')
    ax2.set_xlim(-3, np.percentile(all_vals, 99.98))

    for i in hsim_collect.columns:
        hsim_collect[i].plot(style='.', label='_nolegend_', c='gray', alpha=0.2, ax=ax)
        try:
            hsim_collect.loc[hsim_collect[i] > np.percentile(all_vals, perci), i].plot(style='.', label='_nolegend_',
                                                                                   c='red', alpha=0.2, ax=ax)
        except:
            pass

    (h_count.sum(axis=1) / len(h_count.columns)).rolling(7).mean().plot(ax=ax1, grid=False, ls='--', label='all')
    cc = [x for x in h_count.columns if x.split('.')[-1] == 'rcp45']
    (h_count[cc].sum(axis=1) / len(cc)).rolling(7).mean().plot(ax=ax1, grid=False, label='rcp45')

    cc = [x for x in h_count.columns if x.split('.')[-1] == 'rcp85']
    (h_count[cc].sum(axis=1) / len(cc)).rolling(7).mean().plot(ax=ax1, grid=False, label='rcp85')
    ax1.legend()
    ax.set_ylabel('Q (mm/day)')
    ax1.set_ylabel('days w/ Q>' + str(np.round(perci / 100., 3)) + ' percentile\n7 year rolling mean')
    st.pyplot()


def management_explorer():
    rcps = ['rcp45', 'rcp85']  # 'rcp26',
    rcp = st.sidebar.selectbox('RCP:', rcps)

    data = climate.filter_by_attrs(RCP=rcp)

    ub = st.sidebar.checkbox('Apply SDM bias correction', True)

    cdata = ub_climate(data, weather['krummhoern'], ub)

    if w_topic == 'Agriculture':
        crops = ['wheat', 'maize', 'meadow']
        crop = st.sidebar.selectbox('Crop:', crops)

        yieldx = cropmodel(cdata,crop,rcp,name=crop+' @ '+rcp)

        yieldx.rolling(7).mean().plot()
        st.pyplot()

    elif w_topic == 'Wind Energy':
        storeddata = st.checkbox('Use pre-calculated data (calculation not yet optimized - very slow)', True)
        if storeddata:
            cwind_p = pd.read_csv('data/windpowerx.csv', index_col=0)
            cwind_p.index = pd.to_datetime(cwind_p.index)
            plt.figure(figsize=(10, 2.5))
            cwind_p[[x for x in cwind_p.columns if rcp in x][::3]].mean(axis=1).plot(label='Enercon 3MW')
            cwind_p[[x for x in cwind_p.columns if rcp in x][1::3]].mean(axis=1).plot(label='9x Enercon 0.33MW')
            cwind_p[[x for x in cwind_p.columns if rcp in x][2::3]].mean(axis=1).plot(label='0.4x Enercon 7.5MW')
        else:
            wpengine = ['Enercon 3MW', 'Vestas 3MW', '9x Enercon 0.33MW', '13x Vestas 0.23MW', '0.4x Enercon 7.5MW']
            #wpengine = ['Enercon 3MW', '9x Enercon 0.33MW', '0.4x Enercon 7.5MW']
            turbine = st.sidebar.selectbox('Select turbine:', wpengine)

            plt.figure(figsize=(10,2.5))
            if turbine == 'Enercon 3MW':
                pfi = 'data/pow/Enercon E-115 3000kW (MG).pow'
                P_wind(cdata.u2, pfi).resample('1Y').sum().rolling(7).mean().plot(label=turbine)
            elif turbine == 'Vestas 3MW':
                pfi = 'data/pow/Vestas 112m 3MW (MT).pow'
                P_wind(cdata.u2, pfi).resample('1Y').sum().rolling(7).mean().plot(label=turbine)
            elif turbine == '9x Enercon 0.33MW':
                pfi = 'data/pow/Enercon E33_33.4m _330kw(MT).pow'
                (P_wind(cdata.u2, pfi).resample('1Y').sum()*9.).rolling(7).mean().plot(label=turbine)
            elif turbine == '13x Vestas 0.23MW':
                pfi = 'data/pow/Vestas V29-225kw(MT).pow'
                (P_wind(cdata.u2, pfi).resample('1Y').sum() * 13.).rolling(7).mean().plot(label=turbine)
            elif turbine == '0.4x Vestas 7.5MW':
                pfi = 'data/pow/Enercon E-126_127m_7500kW (MT).pow'
                (P_wind(cdata.u2, pfi).resample('1Y').sum() * 0.4).rolling(7).mean().plot(label=turbine)

        plt.legend()
        plt.ylabel('Wh/yr')
        st.pyplot()

    elif w_topic == 'Extreme Events':
        water_proj()

# main app
## get sub-files
ruinslogo = Image.open('RUINS_logo_small.png')
intro_md = read_markdown_file('explainer/Intro.md')

## sidebar header
st.sidebar.image(ruinslogo, caption='', use_column_width=True)
st.sidebar.header('Control Panel')

## sidebar section navigation
navi_secs = ["Introduction","Weather Data", "Climate Projections", "Uncertainty", "Management"]
navi_sec = st.sidebar.radio("Select Chapter:", options=navi_secs)

topics = ['Warming', 'Weather Indices', 'Drought/Flood', 'Agriculture', 'Extreme Events', 'Wind Energy']
w_topic = st.sidebar.selectbox('Select Topic:', topics)

if navi_sec=="Introduction":
    st.markdown(intro_md, unsafe_allow_html=True)
elif navi_sec=="Weather Data":
    st.header('Weather Data Explorer')
    st.markdown('''In this section we provide visualisations to explore changes in observed weather data. Based on different variables and climate indices it is possible to investigate how climate change manifests itself in different variables, at different stations and with different temporal aggregation.''',unsafe_allow_html=True)
    weather_explorer()
elif navi_sec=="Climate Projections":
    st.header('Climate Projections Explorer')
    st.markdown('''In this section we add climate model projections to the table. The same variables and climate indices are used to explore the projections of different climate models and downscaling models. It is also possible to compare projections under different scenarios about the CO<sub>2</sub>-concentration pathways to observed weather and between different model projections.''',unsafe_allow_html=True)
    climate_explorer()
elif navi_sec=="Uncertainty":
    st.header('Uncertainty Explorer')
elif navi_sec=="Management":
    st.header('Management Explorer')
    st.markdown('''In this section we project the climate model outputs to ecosystem services through diffferent model approaches. The impact of potential management is tackeled at a rather broad level in terms of exemplary options.''', unsafe_allow_html=True)
    management_explorer()



import xarray as xr
import os

from ruins.core import DataManager
from ruins.core.data_manager import HDF5Source

# some datasources are backed by git-lfs which have to be disabled on 
# github actions
NO_LFS = 'NO_LFS' in os.environ


def test_default_manager():
    """Instantiate the default data manager"""
    dm = DataManager()

    assert dm.cache == True

    # find some datasets weather dataset
    assert 'cordex_coast' in dm.datasources
    assert 'CMIP5grid' in dm.datasources


def test_weather_dataset():
    """Test the weather dataset"""
    dm = DataManager()

    # check weather dataset was loaded
    assert 'weather' in dm.datasources

    # check Source type
    weather = dm['weather']
    assert isinstance(weather, HDF5Source)

    # load the data
    if NO_LFS:
        print('No LFS, skipping partial test')
        return

    data = weather.read()
    assert isinstance(data, xr.Dataset)



"""
Topic Selector
==============

"""
from typing import List
import streamlit as st
from PIL import Image

from ruins.core import Config
from ruins.core.debug_view import debug_view


def current_topic_selector(config: Config, expander_container = st.sidebar, **kwargs):
    """
    Select a topic from a list of topics with a short description and a 
    symbolic image.
    """
    container = st if 'container' not in kwargs else kwargs['container']

    if 'topic_list' in kwargs:
        topic_list = kwargs['topic_list']
    else:
        topic_list = config['topic_list']

    # check if the topic was set
    if config.has_key('current_topic'):
        expander_container.selectbox('Select topic:', topic_list, key='current_topic')
        return

    # From here on full topic select
    # introduction
    container.title("Select a topic you would like to explore")
    container.markdown("""Inside the RUINS App you can explore different topics. Please select one, you can select another topic later in the sidebar.""")

    # use a new container for each row
    row_container = st
    # column topic select (image, description, button)
    image_column, description_column, button_column = row_container.columns([1,3,1])
    
    image = Image.open('RUINS_logo_small.png') # idea: display symbol images for each topic
    
    image_column.image(image)
    warming = button_column.button("Click here to select topic Warming")
    description_column.markdown("""**Warming**: In this topic we provide visualisations to explore changes in observed weather data. Based on different variables and climate indices it is possible to investigate how climate change manifests itself in different variables, at different stations and with different temporal aggregation.""")
    row_container.markdown("""___""")

    row_container = st
    # column topic select (image, description, button)
    image_column, description_column, button_column = row_container.columns([1,3,1])

    image_column.image(image)
    weather_indices = button_column.button("Click here to select topic Weather indices")
    description_column.markdown("""**Weather indices**: Short weather indices description.""") # TODO
    row_container.markdown("""___""")
    
    if warming:
        st.session_state.current_topic = 'Warming'  
        st.experimental_rerun()  
    elif weather_indices:
        st.session_state.current_topic = 'Weather Indices'
        st.experimental_rerun()
    else:
        # dev only
        debug_view(dataManager=None, config=config)
        st.stop()



def topic_select(config: Config, expander_container=st.sidebar, **kwargs) -> str:
    """
    Through this function a topic is selected.
    If we are either not in story_mode or if a topic is already selected, 
    a topic can be selected in a selectbox in the sidebar 
    (compact_topic_selector()).
    Otherwise a topic is selected in the full_topic_selector() which 
    provides a list of topics with a short description and a symbolic 
    image.
    """
    # set default story mode 
    story_mode = config.get('story_mode', True)
    if 'story_mode' in kwargs:
        story_mode = kwargs['story_mode']

    # switch the story_mode
    if not story_mode:
        st.session_state.current_topic = 'Warming'
    
    # select topic
    current_topic_selector(config=config, expander_container=expander_container, **kwargs)

    
def debug_main(**kwargs):
    from ruins.core import build_config
    params = st.experimental_get_query_params()
    config, _ = build_config(omit_dataManager=True, url_params=params, **kwargs)

    topic_select(config=config)

    st.json(st.session_state)

if __name__ == '__main__':
    import fire
    fire.Fire(debug_main)



import streamlit as st
import plotly.graph_objs as go
import plotly.express as px
import pandas as pd
import numpy as np

from ruins.core import DataManager
from ruins.core.cache import partial_memoize


@partial_memoize(hash_names=['sel', 'cm'])
def plt_map(dataManager: DataManager, sel='all', cm='none') -> go.Figure:
    # cordex_grid = xr.open_dataset('data/CORDEXgrid.nc')
    # cimp_grid = xr.open_dataset('data/CMIP5grid.nc')
    # stats = pd.read_csv('data/stats.csv', index_col=0)
    cordex_grid = dataManager['CORDEXgrid'].read()
    cimp_grid = dataManager['CMIP5grid'].read()
    stats = dataManager['stats'].read()

    stats['ms'] = 15.
    stats['color'] = 'gray'

    mapbox_access_token = 'pk.eyJ1IjoiY29qYWNrIiwiYSI6IkRTNjV1T2MifQ.EWzL4Qk-VvQoaeJBfE6VSA'
    px.set_mapbox_access_token(mapbox_access_token)

    nodexy = pd.DataFrame([cordex_grid.lon.values.ravel(), cordex_grid.lat.values.ravel()]).T
    nodexy.columns = ['lon', 'lat']
    nodexy['hov'] = 'CORDEX grid'

    def lin_grid(fig, dummy, chex='#d95f0e', opc=0.7):
        x1, x2 = np.shape(dummy.lon)
        lond = np.diff(dummy.lon.values)
        latd = np.diff(dummy.lat.values)
        lond1 = np.diff(dummy.lon.values, axis=0)
        latd1 = np.diff(dummy.lat.values, axis=0)

        for i in np.arange(x2)[1:-1]:
            fig.add_trace(go.Scattermapbox(
                mode='lines',
                lon=dummy.lon.values[:, i] - 0.5 * lond[:, i - 1],
                lat=dummy.lat.values[:, i] - 0.5 * latd[:, i - 1],
                line={'color': chex, 'width': 1},
                hoverinfo='skip',
                opacity=opc))

        fig.add_trace(go.Scattermapbox(
            mode='lines',
            lon=dummy.lon.values[:, i] + 0.5 * lond[:, i - 1],
            lat=dummy.lat.values[:, i] + 0.5 * latd[:, i - 1],
            line={'color': chex, 'width': 1},
            hoverinfo='skip',
            opacity=opc))

        for i in np.arange(x1)[1:-1]:
            fig.add_trace(go.Scattermapbox(
                mode='lines',
                lon=dummy.lon.values[i, :] - 0.5 * lond1[i - 1, :],
                lat=dummy.lat.values[i, :] - 0.5 * latd1[i - 1, :],
                line={'color': chex, 'width': 1},
                hoverinfo='skip',
                opacity=opc))

        fig.add_trace(go.Scattermapbox(
            mode='lines',
            lon=dummy.lon.values[i, :] + 0.5 * lond1[i - 1, :],
            lat=dummy.lat.values[i, :] + 0.5 * latd1[i - 1, :],
            line={'color': chex, 'width': 1},
            hoverinfo='skip',
            opacity=opc))

        return fig

    def add_stats(sel='all'):


        if sel == 'all':
            ac = '#2c7fb8'
            stats1 = stats
        else:
            ac = 'gray'

            fig.add_trace(go.Scattermapbox(
                lat=stats.lat,
                lon=stats.lon,
                mode='markers',
                marker=go.scattermapbox.Marker(
                    size=10,
                    color=ac,
                    opacity=0.8
                ),
                text=stats['Station name']
            ))

            ac = '#df65b0'
            if sel == 'krummhoern':
                stats1 = stats.loc[stats.krummhoern == True]
            elif sel == 'coast':
                stats1 = stats.loc[stats.coast == True]
            elif sel == 'niedersachsen':
                stats1 = stats.loc[stats.niedersachsen == True]
            elif sel == 'inland':
                stats1 = stats.loc[stats.inland == True]
            else:
                try:
                    stats1 = pd.DataFrame(stats.loc[sel]).T
                except:
                    ac = 'gray'
                    stats1 = pd.DataFrame(stats.loc['Norderney']).T

        fig.add_trace(go.Scattermapbox(
            lat=stats1.lat,
            lon=stats1.lon,
            mode='markers',
            marker=go.scattermapbox.Marker(
                size=10,
                color=ac,
                opacity=0.8
            ),
            text=stats1['Station name']
        ))

        return fig

    def add_cmpx(cm):
        maskcordex_coast = [[10, 4], [11, 4], [11, 5], [11, 6], [11, 7], [11, 8], [10, 8], [9, 8], [8, 8], [8, 9],
                            [8, 10], [9, 10], [10, 10], [10, 11], [11, 12], [12, 12], [12, 13], [12, 14], [13, 14],
                            [14, 14], [15, 14], [16, 14], [17, 14], [18, 14], [19, 15], [20, 15], [21, 14], [22, 14],
                            [22, 13],
                            [9, 4], [10, 5], [10, 6], [10, 7], [10, 8], [9, 7], [8, 7], [7, 7], [7, 8], [7, 9], [9, 9],
                            [10, 12], [11, 13], [11, 14], [14, 15], [15, 15], [16, 15], [17, 15], [18, 15], [19, 16],
                            [20, 16], [21, 15], [22, 15],
                            [11, 4], [12, 4], [12, 5], [12, 6], [12, 7], [12, 8], [11, 10], [11, 11], [13, 12],
                            [13, 13], [14, 13], [15, 13], [16, 13], [17, 13], [18, 13], [19, 14], [20, 14], [21, 13]]
        for cc in maskcordex_coast:
            fig.add_trace(go.Scattermapbox(
                lat=[cordex_grid.lat.values[tuple(cc)]],
                lon=[cordex_grid.lon.values[tuple(cc)]],
                mode='markers',
                marker=go.scattermapbox.Marker(
                    size=8,
                    color='#fec44f',
                    opacity=0.8
                ),
                hoverinfo='skip'
            ))
        return fig

    # fig = go.Scattermapbox(lon = nodexy.lon, lat= nodexy.lat, mode='markers', marker_symbol='square', marker_size=15)
    #fig = px.scatter_mapbox(nodexy, lat='lat', lon='lon', center={'lat': 53.0, 'lon': 8.3}, zoom=5, opacity=0.1, hover_data=['hov'])
    fig = px.scatter_mapbox(stats, lat='lat', lon='lon', center={'lat': 53.0, 'lon': 8.6}, zoom=5, size='ms', opacity=0.8, color='color', hover_data=['Station name', 'lat', 'lon'], size_max=10)
    if cm != 'none':
        fig = lin_grid(fig, cordex_grid)
        fig = lin_grid(fig, cimp_grid, '#2c7fb8')
        fig = add_cmpx(cm)
    fig = add_stats(sel)
    fig.update_layout(showlegend=False,width=300, height=350,margin=dict(l=10, r=10, b=10, t=10))  # ,center={'lat':54.0,'lon':8.3}, zoom=7)
    # st.sidebar.plotly_chart(fig)
    return fig




"""
Test the test-suite itself. A dummy streamlit app, which raises an attribute
error should be catched by the test, although not printed to StdErr if run
by streamlit.

For the actual unittests, we import the real functions from the real app.
In the fututre, this test-suite can be extended to test the edge-cases
where streamlit runs into an error, that the plain function does not see.

"""
import pytest


def dummy_app():
    import streamlit as st

    st.title("Dummy App")
    raise AttributeError

def test_dummy_raises():
    with pytest.raises(AttributeError):
        dummy_app()



# BIAS CORRECTION
import numpy as np
import pandas as pd
from scipy.stats import gamma
from scipy.stats import norm
from scipy.signal import detrend

'''
Scaled distribution mapping for climate data

This is a excerpt from pyCAT and the method after Switanek et al. (2017) containing the functions to perform a relative and absolute bias correction on climate data.
(cc) c.jackisch@tu-braunschweig.de, July 2020

It is intended to be used on pandas time series at single locations/pixels.

Switanek, M. B., P. A. Troch, C. L. Castro, A. Leuprecht, H.-I. Chang, R. Mukherjee, and E. M. C. Demaria (2017), Scaled distribution mapping: a bias correction method that preserves raw climate model projected changes, Hydrol. Earth Syst. Sci., 21(6), 2649–2666, https://doi.org/10.5194/hess-21-2649-2017
'''


def relSDM(obs, mod, sce, cdf_threshold=0.9999999, lower_limit=0.1):
    '''relative scaled distribution mapping assuming a gamma distributed parameter (with lower limit zero)
    rewritten from pyCAT for 1D data

    obs :: observed variable time series
    mod :: modelled variable for same time series as obs
    sce :: to unbias modelled time series
    cdf_threshold :: upper and lower threshold of CDF
    lower_limit :: lower limit of data signal (values below will be masked!)

    returns corrected timeseries
    tested with pandas series.
    '''

    obs_r = obs[obs >= lower_limit]
    mod_r = mod[mod >= lower_limit]
    sce_r = sce[sce >= lower_limit]

    obs_fr = 1. * len(obs_r) / len(obs)
    mod_fr = 1. * len(mod_r) / len(mod)
    sce_fr = 1. * len(sce_r) / len(sce)
    sce_argsort = np.argsort(sce)

    obs_gamma = gamma.fit(obs_r, floc=0)
    mod_gamma = gamma.fit(mod_r, floc=0)
    sce_gamma = gamma.fit(sce_r, floc=0)

    obs_cdf = gamma.cdf(np.sort(obs_r), *obs_gamma)
    mod_cdf = gamma.cdf(np.sort(mod_r), *mod_gamma)
    obs_cdf[obs_cdf > cdf_threshold] = cdf_threshold
    mod_cdf[mod_cdf > cdf_threshold] = cdf_threshold

    expected_sce_raindays = min(int(np.round(len(sce) * obs_fr * sce_fr / mod_fr)), len(sce))
    sce_cdf = gamma.cdf(np.sort(sce_r), *sce_gamma)
    sce_cdf[sce_cdf > cdf_threshold] = cdf_threshold

    # interpolate cdf-values for obs and mod to the length of the scenario
    obs_cdf_intpol = np.interp(np.linspace(1, len(obs_r), len(sce_r)), np.linspace(1, len(obs_r), len(obs_r)), obs_cdf)
    mod_cdf_intpol = np.interp(np.linspace(1, len(mod_r), len(sce_r)), np.linspace(1, len(mod_r), len(mod_r)), mod_cdf)

    # adapt the observation cdfs
    obs_inverse = 1. / (1 - obs_cdf_intpol)
    mod_inverse = 1. / (1 - mod_cdf_intpol)
    sce_inverse = 1. / (1 - sce_cdf)
    adapted_cdf = 1 - 1. / (obs_inverse * sce_inverse / mod_inverse)
    adapted_cdf[adapted_cdf < 0.] = 0.

    # correct by adapted observation cdf-values
    xvals = gamma.ppf(np.sort(adapted_cdf), *obs_gamma) * gamma.ppf(sce_cdf, *sce_gamma) / gamma.ppf(sce_cdf,
                                                                                                     *mod_gamma)

    # interpolate to the expected length of future raindays
    correction = np.zeros(len(sce))
    if len(sce_r) > expected_sce_raindays:
        xvals = np.interp(np.linspace(1, len(sce_r), expected_sce_raindays), np.linspace(1, len(sce_r), len(sce_r)),
                          xvals)
    else:
        xvals = np.hstack((np.zeros(expected_sce_raindays - len(sce_r)), xvals))

    correction[sce_argsort[-expected_sce_raindays:]] = xvals

    return pd.Series(correction, index=sce.index)


def absSDM(obs, mod, sce, cdf_threshold=0.9999999):
    '''absolute scaled distribution mapping assuming a normal distributed parameter
    rewritten from pyCAT for 1D data

    obs :: observed variable time series
    mod :: modelled variable for same time series as obs
    sce :: to unbias modelled time series
    cdf_threshold :: upper and lower threshold of CDF

    returns corrected timeseries
    tested with pandas series.
    '''

    obs_len = len(obs)
    mod_len = len(mod)
    sce_len = len(sce)
    obs_mean = np.mean(obs)
    mod_mean = np.mean(mod)
    smean = np.mean(sce)
    odetrend = detrend(obs)
    mdetrend = detrend(mod)
    sdetrend = detrend(sce)

    obs_norm = norm.fit(odetrend)
    mod_norm = norm.fit(mdetrend)
    sce_norm = norm.fit(sdetrend)

    sce_diff = sce - sdetrend
    sce_argsort = np.argsort(sdetrend)

    obs_cdf = norm.cdf(np.sort(odetrend), *obs_norm)
    mod_cdf = norm.cdf(np.sort(mdetrend), *mod_norm)
    sce_cdf = norm.cdf(np.sort(sdetrend), *sce_norm)
    obs_cdf = np.maximum(np.minimum(obs_cdf, cdf_threshold), 1 - cdf_threshold)
    mod_cdf = np.maximum(np.minimum(mod_cdf, cdf_threshold), 1 - cdf_threshold)
    sce_cdf = np.maximum(np.minimum(sce_cdf, cdf_threshold), 1 - cdf_threshold)

    # interpolate cdf-values for obs and mod to the length of the scenario
    obs_cdf_intpol = np.interp(np.linspace(1, obs_len, sce_len), np.linspace(1, obs_len, obs_len), obs_cdf)
    mod_cdf_intpol = np.interp(np.linspace(1, mod_len, sce_len), np.linspace(1, mod_len, mod_len), mod_cdf)

    # adapt the observation cdfs
    # split the tails of the cdfs around the center
    obs_cdf_shift = obs_cdf_intpol - .5
    mod_cdf_shift = mod_cdf_intpol - .5
    sce_cdf_shift = sce_cdf - .5
    obs_inverse = 1. / (.5 - np.abs(obs_cdf_shift))
    mod_inverse = 1. / (.5 - np.abs(mod_cdf_shift))
    sce_inverse = 1. / (.5 - np.abs(sce_cdf_shift))
    adapted_cdf = np.sign(obs_cdf_shift) * (1. - 1. / (obs_inverse * sce_inverse / mod_inverse))
    adapted_cdf[adapted_cdf < 0] += 1.
    adapted_cdf = np.maximum(np.minimum(adapted_cdf, cdf_threshold), 1 - cdf_threshold)

    xvals = norm.ppf(np.sort(adapted_cdf), *obs_norm) \
            + obs_norm[-1] / mod_norm[-1] \
            * (norm.ppf(sce_cdf, *sce_norm) - norm.ppf(sce_cdf, *mod_norm))
    xvals -= xvals.mean()
    xvals += obs_mean + (smean - mod_mean)

    correction = np.zeros(sce_len)
    correction[sce_argsort] = xvals
    correction += sce_diff - smean

    return correction


def SDM(obs, mod, sce, meth='rel', cdf_threshold=0.9999999, lower_limit=0.1):
    '''scaled distribution mapping - wrapper to relative and absolute bias correction functions
    rewritten from pyCAT for 1D data

    obs :: observed variable time series
    mod :: modelled variable for same time series as obs
    sce :: to unbias modelled time series
    meth :: 'rel' for relative SDM, else absolute SDM will be performed
    cdf_threshold :: upper and lower threshold of CDF
    lower_limit :: lower limit of data signal (values below will be masked when meth != 'rel')

    The original authors suggest to use the absolute SDM for air temperature and the relative SDM for precipitation and radiation series.

    returns corrected timeseries
    tested with pandas series.
    '''

    if meth == 'rel':
        return relSDM(obs, mod, sce, cdf_threshold, lower_limit)
    else:
        return absSDM(obs, mod, sce, cdf_threshold)


from typing import List, Callable, Tuple
import streamlit as st
import xarray as xr     # TODO: these references should be moved to DataManager
import pandas as pd     # TODO: these references should be moved to DataManager
import numpy as np
import matplotlib.pyplot as plt

from ruins.plotting import plt_map, kde, yrplot_hm
from ruins.components import data_select, topic_select
from ruins.core import build_config, debug_view, DataManager, Config
from ruins.core.cache import partial_memoize



####
# OLD STUFF
def applySDM(wdata, data, meth='rel', cdf_threshold=0.9999999, lower_limit=0.1):
    '''apply structured distribution mapping to climate data and return unbiased version of dataset'''
    from sdm import SDM
    data_ub = data.copy()

    for k in data_ub.columns:
        data_col = data_ub[k].dropna()
        overlapx = pd.concat(
            [wdata.loc[data_col.index[0]:wdata.index[-1]], data_col.loc[data_col.index[0]:wdata.index[-1]]], axis=1)
        overlapx.columns = ['obs', 'cm']
        overlapx = overlapx.dropna()
        try:
            data_ub[k] = SDM(overlapx.obs, overlapx.cm, data_col, meth, cdf_threshold, lower_limit)
        except:
            data_ub[k] = data_ub[k] * np.nan

    data_ub[data_ub == 0.0000000] = np.nan
    data_ub = data_ub.loc[data_ub.index[0]:pd.to_datetime('2099-12-31 23:59:59')]

    return data_ub


def climate_indi(ts, indi='Summer days (Tmax ≥ 25°C)'):
    '''
    Calculate climate indicator days.
    Input time series of meteorological data
    '''
    if pd.infer_freq(ts.index) != 'D':
        print('Please provide daily data.')
        return

    # drop NA
    ts = ts.dropna()

    if indi == 'Summer days (Tmax ≥ 25°C)':  # summer days
        return (ts.Tmax >= 25.).groupby(ts.index.year).sum()
    elif indi == 'Ice days (Tmax < 0°C)':  # ice days
        return (ts.Tmax < 0.).groupby(ts.index.year).sum()
    elif indi == 'Frost days (Tmin < 0°C)':  # frost days
        return (ts.Tmin < 0.).groupby(ts.index.year).sum()
    elif indi == 'Hot days (Tmax ≥ 30°C)':  # hot days
        return (ts.Tmax >= 30.).groupby(ts.index.year).sum()
    elif indi == 'Tropic nights (Tmin ≥ 20°C)':  # tropic night
        return (ts.Tmin >= 20.).groupby(ts.index.year).sum()
    elif indi == 'Rainy days (Precip ≥ 1mm)':  # rainy days
        return (ts.Prec >= 1.).groupby(ts.index.year).sum()
    else:
        print('Nothing calculated.')
        return

# TODO: document + signature
# TODO: extract plotting
def climate_indices(dataManager: DataManager, config: Config):
    # get data
    weather = dataManager['weather'].read()
    climate = dataManager['cordex_coast'].read()

    # get the relevant settings
    stati = config.get('selected_station', 'coast')


    cindi = ['Ice days (Tmax < 0°C)', 'Frost days (Tmin < 0°C)', 'Summer days (Tmax ≥ 25°C)', 'Hot days (Tmax ≥ 30°C)','Tropic nights (Tmin ≥ 20°C)', 'Rainy days (Precip ≥ 1mm)']
    ci_topic = st.selectbox('Select Index:', cindi)

    if ci_topic == 'Rainy days (Precip ≥ 1mm)':
        vari = 'Prec'
        meth = 'rel'
    elif (ci_topic == 'Frost days (Tmin < 0°C)') | (ci_topic == 'Tropic nights (Tmin ≥ 20°C)'):
        vari = 'Tmin'
        meth = 'abs'
    else:
        vari = 'Tmax'
        meth = 'abs'

    w1 = weather[stati].sel(vars=vari).to_dataframe()
    w1.columns = ['bla', vari]

    fig = plt.figure(figsize=(10,2.5))
    wi = climate_indi(w1, ci_topic).astype(int)
    wi.plot(style='.', color='steelblue', label='Coast weather')
    wi.rolling(10, center=True).mean().plot(color='steelblue', label='Rolling mean\n(10 years)')

    if config['include_climate']:
        c1 = climate.sel(vars=vari).to_dataframe()
        c1 = c1[c1.columns[c1.columns != 'vars']]
        c2 = applySDM(w1[vari], c1, meth=meth)

        firstitem = True
        for j in c2.columns:
            dummyt = pd.DataFrame(c2[j])
            dummyt.columns = [vari]
            ci = pd.DataFrame(climate_indi(dummyt, ci_topic).astype(int))
            ci.columns = [j]

            if firstitem:
                ci[j].plot(style='.', color='gray', alpha=0.1, label='all climate projections')
                cid = ci
                firstitem = False
            else:
                ci[j].plot(style='.', color='gray', alpha=0.1, label='_')
                cid = pd.concat([cid, ci], axis=1)

        rcpx = np.array([x[-5:] for x in cid.columns.values])
        for n in np.unique(rcpx):
            nx = n
            if n == np.unique(rcpx)[-1]:
                nx = n + '\n(Rolling of 5 years)'
            cid[cid.columns[rcpx == n]].mean(axis=1).rolling(5, center=True).mean().plot(label='Mean of ' + nx)

    plt.legend(ncol=2)
    plt.ylabel('Number of days')
    plt.title(ci_topic)
    st.pyplot(fig)

    if ci_topic == 'Ice days (Tmax < 0°C)':
        st.markdown('''Number of days in one year which persistently remain below 0°C air temperature.''')
    elif ci_topic == 'Frost days (Tmin < 0°C)':
        st.markdown('''Number of days in one year which reached below 0°C air temperature.''')
    elif ci_topic == 'Summer days (Tmax ≥ 25°C)':
        st.markdown('''Number of days in one year which reached or exceeded 25°C air temperature.''')
    elif ci_topic == 'Hot days (Tmax ≥ 30°C)':
        st.markdown('''Number of days in one year which reached or exceeded 30°C air temperature.''')
    elif ci_topic == 'Tropic nights (Tmin ≥ 20°C)':
        st.markdown('''Number of days in one year which persistently remained above 20°C air temperature.''')
    elif ci_topic == 'Rainy days (Precip ≥ 1mm)':
        st.markdown('''Number of days in one year which received at least 1 mm precipitation.''')
    return


@partial_memoize(hash_names=['name', 'station', 'variable', 'time', '_filter'])
def _reduce_weather_data(dataManager: DataManager, name: str, variable: str, time: str, station: str = None, _filter: dict = None) -> pd.DataFrame:
    # get weather data
    arr: xr.Dataset = dataManager[name].read()

    if _filter is not None:
        arr = arr.filter_by_attrs(**_filter)

    if station is None:
        base = arr
    else:
        base = arr[station]

    # reduce to station and variable
    reduced = base.sel(vars=variable).resample(time=time)

    if variable == 'Tmax':
        df = reduced.max(dim='time').to_dataframe()
    elif variable == 'Tmin':
        df = reduced.min(dim='time').to_dataframe()
    else:
        df = reduced.mean(dim='time').to_dataframe()
    
    if station is None:
        return df.loc[:, df.columns != 'vars']
    else:
        return df[station]       


def warming_data_plotter(dataManager: DataManager, config: Config):
    weather: xr.Dataset = dataManager['weather'].read()
    climate = dataManager['cordex_coast'].read()
    statios = list(weather.keys())
    stat1 = config['selected_station']

    # build the placeholders
    plot_area = st.container()
    control_left, control_right = st.columns((1, 3))

    # TODO refactor in data-aggregator and data-plotter for different time frames


    # ----
    # data-aggregator controls
    navi_vars = ['Maximum Air Temperature', 'Mean Air Temperature', 'Minimum Air Temperature']
    navi_var = control_left.radio("Select variable:", options=navi_vars)
    if navi_var[:4] == 'Mini':
        vari = 'Tmin'
        ag = 'min'
    elif navi_var[:4] == 'Maxi':
        vari = 'Tmax'
        ag = 'max'
    else:
        vari = 'T'
        ag = 'mean'

    # controls end
    # ----

    # TODO: this produces a slider but also needs some data caching
    if config['temporal_agg'] == 'Annual':
        wdata = _reduce_weather_data(dataManager, name='weather', station=config['selected_station'], variable=vari, time='1Y')
        allw = _reduce_weather_data(dataManager, name='weather', variable=vari, time='1Y')

        dataLq = float(np.floor(allw.min().quantile(0.22)))
        datamin = float(np.min([dataLq, np.round(allw.min().min(), 1)]))
        
        if config['include_climate']:
            # get the rcp
            rcp = config['current_rcp']

            data = _reduce_weather_data(dataManager, name='cordex_coast', variable=vari, time='1Y', _filter=dict(RCP=rcp))
            data_ub = applySDM(wdata, data, meth='abs')

            dataUq = float(np.ceil(data_ub.max().quantile(0.76)))
            datamax = float(np.max([dataUq, np.round(data_ub.max().max(), 1)]))
        else:
            dataUq = float(np.ceil(allw.max().quantile(0.76)))
            datamax = float(np.max([dataUq,np.round(allw.max().max(), 1)]))

        datarng = control_right.slider('Adjust data range on x-axis of plot:', min_value=datamin, max_value=datamax, value=(dataLq, dataUq), step=0.1, key='drangew')

        # -------------------
        # start plotting plot
        if config['include_climate']:
            fig, ax = kde(wdata, data_ub.mean(axis=1), split_ts=3)
        else:
            fig, ax = kde(wdata, split_ts=3)

        ax.set_title(stat1 + ' Annual ' + navi_var)
        ax.set_xlabel('T (°C)')
        ax.set_xlim(datarng[0],datarng[1])
        plot_area.pyplot(fig)

        sndstat = st.checkbox('Show second station for comparison')

        if sndstat:
            stat2 = st.selectbox('Select second station:', [x for x in statios if x != config['selected_station']])
            wdata2 = _reduce_weather_data(dataManager, name='weather', station=stat2, variable=vari, time='1Y')

            fig, ax2 = kde(wdata2, split_ts=3)
            ax2.set_title(stat2 + ' Annual ' + navi_var)
            ax2.set_xlabel('T (°C)')
            ax2.set_xlim(datarng[0],datarng[1])
            st.pyplot(fig)

        # Re-implement this as a application wide service
        # expl_md = read_markdown_file('explainer/stripes.md')
        # st.markdown(expl_md, unsafe_allow_html=True)

    elif config['temporal_agg'] == 'Monthly':
        wdata = _reduce_weather_data(dataManager, name='weather', station=config['selected_station'], variable=vari, time='1M')

        ref_yr = control_right.slider('Reference period for anomaly calculation:', min_value=int(wdata.index.year.min()), max_value=2020,value=(max(1980, int(wdata.index.year.min())), 2000))

        if config['include_climate']:
            # get the rcp
            rcp = config['current_rcp']
            data = _reduce_weather_data(dataManager, name='cordex_coast', variable=vari, time='1M', _filter=dict(RCP=rcp))

            #ub = st.sidebar.checkbox('Apply SDM bias correction',True)
            ub = True # simplify here and automatically apply bias correction

            if ub:
                data_ub = applySDM(wdata, data, meth='abs')
                fig = yrplot_hm(pd.concat([wdata.loc[wdata.index[0]:data.index[0] - pd.Timedelta('1M')], data_ub.mean(axis=1)]), ref_yr, ag, li=2006, lang=config.lang)
            else:
                fig = yrplot_hm(pd.concat([wdata.loc[wdata.index[0]:data.index[0] - pd.Timedelta('1M')], data.mean(axis=1)]), ref_yr, ag, li=2006, lang=config.lang)

            # old matplotlib include
            #plt.title(stat1 + ' ' + navi_var + ' anomaly to ' + str(ref_yr[0]) + '-' + str(ref_yr[1]))
            #plot_area.pyplot(fig)
            fig.update_layout(title = f"{stat1} {navi_var} anomaly to {ref_yr[0]}-{ref_yr[1]}")
            plot_area.plotly_chart(fig, use_container_width=True)

            # compare to second station
            sndstat = st.checkbox('Compare to a second station?')
            
            if sndstat:
                stat2 = st.selectbox('Select second station:', [x for x in statios if x != stat1])
                wdata2 = _reduce_weather_data(dataManager, name='weather', station=stat2, variable=vari, time='1M')

                #ub = st.sidebar.checkbox('Apply SDM bias correction',True)
                ub = True # simplify here and automatically apply bias correction

                if ub:
                    data2_ub = applySDM(wdata2, data, meth='abs')
                    fig = yrplot_hm(pd.concat([wdata2.loc[wdata2.index[0]:data.index[0] - pd.Timedelta('1M')], data2_ub.mean(axis=1)]), ref_yr, ag, li=2006, lang=config.lang)
                else:
                    fig = yrplot_hm(pd.concat([wdata2.loc[wdata2.index[0]:data.index[0] - pd.Timedelta('1M')], data.mean(axis=1)]), ref_yr, ag, li=2006, lang=config.lang)
                
                fig.update_layout(title=f"{stat2} {navi_var} anomaly to {ref_yr[0]}-{ref_yr[1]}")
                plot_area.plotly_chart(fig, use_container_width=True)

        # TODO: break up this as well
        else:
            fig = yrplot_hm(sr=wdata, ref=ref_yr, ag=ag, lang=config.lang)
            fig.update_layout(title = f"{stat1} {navi_var} anomaly to {ref_yr[0]}-{ref_yr[1]}")
            plot_area.plotly_chart(fig, use_container_width=True)

            # old matplotlib include
            #plt.title(stat1 + ' ' + navi_var + ' anomaly to ' + str(ref_yr[0]) + '-' + str(ref_yr[1]))
            #plot_area.pyplot(fig)

            sndstat = st.checkbox('Compare to a second station?')

            if sndstat:
                stat2 = st.selectbox('Select second station:', [x for x in statios if x != stat1])
                data2 = _reduce_weather_data(dataManager, name='weather', station=stat2, variable=vari, time='1M')

                ref_yr2 = list(ref_yr)
                if ref_yr2[1]<data2.index.year.min():
                    ref_yr2[0] = data2.index.year.min()
                    ref_yr2[1] = ref_yr2[0]+10
                if ref_yr2[0]<data2.index.year.min():
                    ref_yr2[0] = data2.index.year.min()
                    if ref_yr2[1] - ref_yr2[0] < 10:
                        ref_yr2[1] = ref_yr2[0] + 10

                fig = yrplot_hm(sr=data2, ref=ref_yr2, ag=ag, lang=config.lang)
                fig.update_layout(title=f"{stat2} {navi_var} anomaly to {ref_yr2[0]}-{ref_yr2[1]}")
                plot_area.plotly_chart(fig, use_container_width=True)
                
                # old matplotlib include
                #plt.title(stat2 + ' ' + navi_var + ' anomaly to ' + str(ref_yr2[0]) + '-' + str(ref_yr2[1]))
                #st.pyplot(fig)

        # Re-implement this as a application wide service
        # expl_md = read_markdown_file('explainer/stripes_m.md')
        # st.markdown(expl_md, unsafe_allow_html=True)


def weather_explorer(config: Config, dataManager: DataManager):
    """
    TODO: refactor this whole app into the main_app
    """
    # switch the topic
    topic = config['current_topic']
    if topic == 'Warming':
        warming_data_plotter(dataManager, config)
    
    elif topic == 'Weather Indices':
        climate_indices(dataManager, config)



def main_app(**kwargs):
    """Describe the params in kwargs here
    """
    # build the config and dataManager from kwargs
    url_params = st.experimental_get_query_params()
    config, dataManager = build_config(url_params=url_params, **kwargs)

    # set page properties and debug view    
    st.set_page_config(page_title='Weather Explorer', layout=config.layout)
    debug_view.debug_view(dataManager, config, debug_name='DEBUG - initial state')

    # Story mode - go through each setting
    # update session state with current data settings
    data_expander = st.sidebar.expander('Data selection', expanded=True)
    data_select.data_select(dataManager, config, expander_container=data_expander, container=st)
    
    # topic selector
    topic_expander = st.sidebar.expander('Topic selection', expanded=True) # here we would use control_policy ('show', 'hide) if we want to keep it
    topic_select.topic_select(config=config, expander_container=topic_expander, container=st)
    
    # build the app
    st.header('Weather Data Explorer')
    # TODO: move this text also into story mode?
    st.markdown('''In this section we provide visualisations to explore changes in observed weather data. Based on different variables and climate indices it is possible to investigate how climate change manifests itself in different variables, at different stations and with different temporal aggregation.''',unsafe_allow_html=True)

    # topic is set, now run the correct plot
    topic = config['current_topic']

    if topic == 'Warming':
        warming_data_plotter(dataManager, config)
    
    elif topic == 'Weather Indices':
        climate_indices(dataManager, config)

    else:
        st.error(f'{topic} is not a valid topic here. Please tell the developer.')

    # end state debug
    debug_view.debug_view(dataManager, config, debug_name='DEBUG - finished app')


if __name__ == '__main__':
    import fire
    fire.Fire(main_app)



import importlib
from setuptools import setup, find_packages


def readme():
    with open('README.md') as f:
        return f.read()


def requirements():
    with open('requirements.txt') as f:
        return f.read().splitlines()


def version():
    ruins = importlib.import_module('ruins')
    return ruins.__version__


setup(
    name='ruins-app',
    description='Climate change and uncertainty explorer',
    long_description=readme(),
    version=version(),
    author='Conrad Jackisch, Mirko Mälicke',
    author_email='Conrad.Jackisch@tbt.tu-freiberg.de, mirko@hydrocode.de',
    install_requires=requirements(),
    packages=find_packages()
)




# GR4J hydrological bucket model after Fabrizo Fenicia

import numpy as np
from numba import jit

# HELPER FUNCTIONS:
def SS1(i,tBase):
    if (i<=0):
        y=0
    elif (i<tBase):
        y=(i/tBase)**2.5
    else:
        y=1
    return y

def SS2(i,tBase):
    if (i<=0):
        y=0
    elif (i<tBase):
        y=0.5*(i/tBase)**2.5
    elif (i < 2*tBase):
        y = 1-0.5 * (2-i / tBase) ** 2.5
    else:
        y=1
    return y

def tWeightsL1(tBase):
    iMax = int(np.ceil(tBase))
    w = np.zeros((iMax,1))
    for i in np.arange(iMax):
        w[i]=SS1(i,tBase)-SS1(i-1,tBase)
    return w

def tWeightsL2(tBase):
    iMax = int(np.ceil(tBase*2))
    w = np.zeros((iMax,1))
    for i in np.arange(iMax):
        w[i]=SS2(i,tBase)-SS2(i-1,tBase);
    return w

# STORAGE FUNCTIONS
def fun_Qq_UR(S,Pn,x1):
    return (x1*(1-(S/x1)**2)*np.tanh(Pn/x1))/(1+S/x1*np.tanh(Pn/x1))

def fun_Qp_UR(S,x1):
    return S*(1-(1+(4/9.*S/x1)**4)**-0.25)

def fun_E_UR(S,Ep,x1):
    return (S*(2-S/x1)*np.tanh(Ep/x1))/(1+(1-S/x1)*np.tanh(Ep/x1))



def pyGR4J(data,para,w1,w2,start_i=-99,nT=-99):
    # forcings
    P = data[:,0]
    Epot = data[:,1]

    par_SiniFr_UR = 0.
    par_x1,par_x2,par_x3,par_x4,par_spl = para

    if start_i == -99:
        start_i = 0
    if nT == -99:
        nT = len(data) - start_i

    #allocate reservoirs
    st_UR = np.zeros(nT)*np.nan   #production store
    st_RR = np.zeros(nT) * np.nan #routing store

    lW1 = len(w1)
    lW2 = len(w2)
    stLag1 = np.zeros(lW1) # store lag function values
    stLag2 = np.zeros(lW2) # store lag function values

    #fluxes
    Eact = np.zeros(nT)*np.nan
    Qsim = np.zeros(nT)*np.nan

    #initialize states
    stStart_UR = par_SiniFr_UR * par_x1
    stStart_RR = 0.

    st_UR[0] = stStart_UR
    st_RR[0] = stStart_RR

    ###
    # RUN MODEL
    for i in np.arange(nT):
        # interception
        E_IR = np.min([Epot[i], P[i]])
        E_URpot = Epot[i] - E_IR
        P_UR = P[i] - E_IR
        # production store
        Qq_UR = P_UR - fun_Qq_UR(st_UR[i], P_UR, par_x1)
        E_UR = fun_E_UR(st_UR[i], E_URpot, par_x1)
        st_UR[i] = st_UR[i] + P_UR - Qq_UR - E_UR
        Eact[i] = E_IR + E_UR
        # percolation
        Qp_UR = fun_Qp_UR(st_UR[i], par_x1)
        st_UR[i] = st_UR[i] - Qp_UR
        # split
        P_SP = Qp_UR + Qq_UR
        P_L1 = par_spl * P_SP
        P_L2 = P_SP - P_L1
        # lag 1
        for k in np.arange(lW1 - 1):
            stLag1[k] = stLag1[k + 1] + w1[k] * P_L1
        stLag1[lW1 - 1] = w1[lW1 - 1] * P_L1
        # lag 2
        for k in np.arange(lW2 - 1):
            stLag2[k] = stLag2[k + 1] + w2[k] * P_L2
        stLag2[lW2 - 1] = w2[lW2 - 1] * P_L2
        # Water exchange
        Q_EX = par_x2 * (st_RR[i] / par_x3) ** 3.5
        # Routing store
        st_RR[i] = np.max([0., st_RR[i] + stLag1[0] + Q_EX])
        R2 = st_RR[i] / (1 + (st_RR[i] / par_x3) ** 4) ** 0.25
        Q_RR = st_RR[i] - R2
        st_RR[i] = R2
        # Output
        QD = np.max([0., stLag2[0] + Q_EX])
        Qsim[i] = Q_RR + QD
        if (i < nT - 1):
            st_UR[i + 1] = st_UR[i]
            st_RR[i + 1] = st_RR[i]

    return Qsim